from glypy.composition.ccomposition cimport CComposition

cdef class _CompositionBase(dict):
    cdef:
        public object _mass
        public object _reducing_end
        public CComposition _composition_offset

    cpdef object _getitem_fast(self, object key)
    cpdef object _setitem_fast(self, object key, object value)

    cpdef object _update_from_typed_map(self, _CompositionBase template, bint copy_nodes=*)
    cpdef str serialize(self)

    cdef void _add_from(self, _CompositionBase other)
    cdef void _subtract_from(self, _CompositionBase other)


'''
Represent a sugar graph with pseudo-directed edges.

'''
import warnings
import operator
import logging
import itertools
from functools import partial
from collections import deque, defaultdict
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable

from glypy.utils import (
    identity,
    chrinc,
    uid)
from glypy.composition import Composition

from .base import SaccharideCollection
from .monosaccharide import (
    Monosaccharide,
    graph_clone,
    toggle as residue_toggle,
    MonosaccharideOccupancy)
from .constants import UnknownPosition, NoPosition
from .substituent import Substituent
from .crossring_fragments import crossring_fragments, CrossRingPair
from .fragment import Subtree

logger = logging.getLogger("Glycan")

_fragment_direction = {
    "A": -1,
    "B": -1,
    "C": -1,
    "X": 1,
    "Y": 1,
    "Z": 1
}

MAIN_BRANCH_SYM = '-'


def fragment_to_substructure(fragment, tree):
    """Extract the substructure of `tree` which is contained in `fragment`

    Parameters
    ----------
    fragment: :class:`~.GlycanFragment`
        The :class:`~.GlycanFragment` to extract substructure for.
    tree: :class:`Glycan`
        The |Glycan| to extract substructure from.

    Returns
    -------
    :class:`Glycan`:
        The |Glycan| substructure defined by the nodes contained in `fragment` as
        found in `tree`
    """
    break_targets = fragment.link_ids
    crossring_targets = fragment.crossring_cleavages

    # All operations will be done on a copy of the tree of interest
    tree = tree.clone()

    crossring_targets_nodes = []
    break_targets_nodes = []
    # A point of reference known to be inside the fragment tree
    anchor = None
    for pos, link in tree.iterlinks():
        # If the current link's child was cross-ring cleaved,
        # then we must apply that cleavage and find an anchor
        if link.child.id in crossring_targets:
            ion_type = crossring_targets[link.child.id]
            c1, c2 = map(int, ion_type[0].split(","))
            target = tree.get(link.child.id)
            crossring_targets_nodes.append((target, c1, c2))
        # If this link was cleaved, break it and find an anchor
        if link.id in break_targets:
            break_targets_nodes.append(link)

    for target, c1, c2 in crossring_targets_nodes:
        a_frag, x_frag = crossring_fragments(
            target, c1, c2, attach=True, copy=False)
        next(residue_toggle(target))

        if crossring_targets[target.id][1] == "A":
            anchor = a_frag
        else:
            anchor = x_frag

    for link in break_targets_nodes:
        parent, child = link.break_link(refund=True)
        if parent.id in fragment.included_nodes:
            anchor = parent
        elif child.id in fragment.included_nodes:
            anchor = child

    # Build a new tree from the anchor
    substructure = Glycan(root=anchor, index_method=None).reroot()

    return substructure


class Glycan(SaccharideCollection):
    '''
    Represents a full graph of connected |Monosaccharide| objects and their connecting bonds.

    Attributes
    ----------
    root: |Monosaccharide|
        The first monosaccharide unit of the glycan, and the reducing end if present.
    index: |list|
        A list of the |Monosaccharide| instances in `self` in the order they are encountered
        by traversal by `traversal_methods[index_method]`
    link_index: |list|
        A list of the |Link| connecting the |Monosaccharide| instances in `self` in the order they
        are encountered by traversal by `traversal_methods[index_method]`
    reducing_end: |ReducedEnd| or |None|
        The reducing end on :attr:`root`.
    branch_lengths: |dict|
        A dictionary mapping branch symbols to their lengths
    branch_parent_map: |dict|
        A dictionary mapping branch symbols to their parent branch symbols
    '''

    verbose = False

    _serializers = {}

    @classmethod
    def subtree_from(cls, tree, node, visited=None):
        if isinstance(node, int):
            node = tree[node]
        visited = {
            node.id for p,
            node in node.parents()} if visited is None else visited
        subtree = cls(root=node, index_method=None).clone(
            index_method=None, visited=visited)
        return subtree

    def fragment_to_substructure(self, fragment):
        """Extract the substructure of `tree` which is contained in `fragment`

        Parameters
        ----------
        fragment: :class:`~.GlycanFragment`
            The :class:`~.GlycanFragment` to extract substructure for.
        tree: :class:`Glycan`
            The |Glycan| to extract substructure from.

        Returns
        -------
        :class:`Glycan`:
            The |Glycan| substructure defined by the nodes contained in `fragment` as
            found in `tree`
        """
        return fragment_to_substructure(fragment, self)

    traversal_methods = {}

    def __init__(self, root=None, index_method='dfs', canonicalize=False):
        '''
        Constructs a new :class:`Glycan` from the collection of connected :class:`~.Monosaccharide`
        objects rooted at `root`.

        If :obj:`index_method` is not |None|, the graph is indexed by the default search method
        given by :obj:`traversal_methods[index_method]`
        '''
        if root is None:
            root = Monosaccharide()
        self.root = root
        self.index = []
        self.link_index = []
        self.branch_lengths = {}
        if index_method is not None:
            self.reindex(method=index_method)
        if canonicalize:
            self.canonicalize()

    def has_index(self):
        return bool(self.index)

    def reindex(self, method='dfs'):
        '''
        Traverse the graph using the function specified by `method`. The order of
        traversal defines the new :attr:`id` value for each |Monosaccharide|
        and |Link|.

        The order of traversal also defines the ordering of the |Monosaccharide|
        in :attr:`index` and |Link| in :attr:`link_index`.

        Prior to constructing a |Glycan| instance, component |Monosaccharide| instances
        may be labeled, converting their id field into a tuple.

        Calls :meth:`label_branches` after indexing is complete.

        Returns
        -------
        Glycan:
            self

        See Also
        --------
        deindex
        label_branches
        '''
        self.deindex()
        traversal = self._get_traversal_method(method)
        index = []
        visited = set()
        i = 1
        for node in traversal():
            addr = id(node)
            if addr in visited:
                continue
            visited.add(addr)
            index.append(node)

        for node in index:
            node.id = i
            i += 1
            # reindex substituents as well
            try:
                for j, subst in node.substituents():
                    subst.id = i
                    i += 1
            except AttributeError:
                if node.node_type is Substituent.node_type:
                    continue
        link_index = []
        for pos, link in self.iterlinks(method=method):
            link_index.append(link)

        i = 1
        for link in link_index:
            link.id = i
            i += 1

        self.index = index
        self.link_index = link_index

        self.label_branches()
        return self

    def _build_link_index(self, method='dfs'):
        link_index = []
        for pos, link in self.iterlinks(method=method):
            link_index.append(link)
        self.link_index = link_index

    def _build_node_index(self, method='dfs'):
        index = []
        for node in self.iternodes(method=method):
            index.append(node)
        self.index = index

    def deindex(self):
        '''
        When combining two Glycan structures, very often their component ids will
        overlap, making it impossible to differentiate between a cycle and the new
        graph. This function mangles all of the node and link ids so that they are
        distinct from the pre-existing nodes.

        Returns
        -------
        Glycan:
            self
        '''
        if self.index is not None and len(self.index) > 0:
            base = uid()
            for node in self.index:
                node.id += base
                node.id *= -1
                try:
                    for j, subst in node.substituents():
                        subst.id += base
                        subst.id *= -1
                except AttributeError:
                    if node.node_type is Substituent.node_type:
                        warnings.warn("A substituent has been detected as a parent "
                                      "node. This glycan may not fragment correctly")
            for link in self.link_index:
                link.id += base
                link.id *= -1
        return self

    def reroot(self, index_method='dfs'):
        """Set :attr:`root` to the node with the lowest :attr:`id`.

        Should only be used if the glycan has been indexed.

        Parameters
        ----------
        index_method : str, optional
            The name of the index method to use to reindex the glycan relative to
            the new root node. If :const:`None`, no reindexing is done. The default
            is 'dfs'

        Returns
        -------
        Glycan:
            self
        """
        self.root = sorted(iter(self), key=operator.attrgetter('id'))[0]
        if index_method is not None:
            self.reindex(method=index_method)
        return self

    def initialize_structure(self, method='dfs'):
        """Rebuild the structure index from scratch using :meth:`reindex`,
        then sets the traversal order using :meth:`canonicalize`.

        Use this method when building a :class:`Glycan` instance from a manually
        created :class:`~.Monosaccharide` graph.
        """
        self.canonicalize()
        self.reindex(method=method)
        return self

    def __getitem__(self, ix):
        '''
        Fetch a :class:`~.Monosaccharide` from :attr:`index`.

        Returns
        -------
        Monosaccharide

        Raises
        ------
        IndexError:
            If the provided `ix` exceeds the length of the index,
            or if :attr:`index` has not been populated.
        '''
        if self.index is not None:
            return self.index[ix]
        else:
            raise IndexError(
                "Tried to access the index of an unindexed Glycan.")

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __root__(self):
        return self.root

    def __tree__(self):
        return self

    def get(self, ix):
        """Get a :class:`~.Monosaccharide` from this structure by its :attr:`~.Monosaccharide.id` value.

        If :attr:`index` is populated it will be iterated over, otherwise :meth:`__iter__` will be called.

        Parameters
        ----------
        ix : int or tuple
            The id value to search for.

        Returns
        -------
        Monosaccharide

        Raises
        ------
        IndexError:
            If the id value is not found
        """
        if self.index:
            iterable = self.index
        else:
            iterable = self
        for node in iterable:
            if node.id == ix:
                return node
        raise IndexError(
            "Could not find a node with the given id {}".format(ix))

    def get_link(self, ix):
        """Search for a :class:`~.Link` by :attr:`~.Link.id` value.

        This will use :meth:`iterlinks` to iterate over the linkages
        in the structure

        Parameters
        ----------
        ix : int
            The link index to search for.

        Returns
        -------
        :class:`~.Link`

        Raises
        ------
        IndexError:
            if the id value is not found
        """
        for pos, link in self.iterlinks():
            if link.id == ix or link.label == ix:
                return link
        raise IndexError(
            "Could not find a link with the given id or label {}".format(ix))

    @property
    def reducing_end(self):
        '''
        An alias for :attr:`Monosaccharide.reducing_end` for :attr:`root`
        '''
        try:
            return self.root.reducing_end
        except AttributeError:
            if self.root.node_type is Substituent.node_type:
                return None
            else:
                raise

    def set_reducing_end(self, value):
        '''
        Sets the reducing end type, and configures the :attr:`root` |Monosaccharide| appropriately.

        If the reducing_end is not |None|, then the following state changes are made to :attr:`root`:

        .. code-block:: python

            self.root.ring_start = 0
            self.root.ring_end = 0
            self.root.anomer = "uncyclized"

        Else, the correct state is unknown:

        .. code-block:: python

            self.root.ring_start = UnknownPosition
            self.root.ring_end = UnknownPosition
            self.root.anomer = None

        .. note:: This method is called automatically when setting :attr:`reducing_end`, and does not

        need to be used explicitly.
        '''
        if self.root.node_type is Substituent.node_type:
            raise TypeError("Cannot set reducing end on this glycan composition")
        self.root.reducing_end = value
        if self.reducing_end is not None:
            self.root.ring_start = 0
            self.root.ring_end = 0
            self.root.anomer = "uncyclized"
        else:
            self.root.ring_start = UnknownPosition
            self.root.ring_end = UnknownPosition
            self.root.anomer = None

    @reducing_end.setter
    def reducing_end(self, value):
        self.set_reducing_end(value)

    def depth_first_traversal(self, from_node=None, apply_fn=identity, visited=None):
        '''
        Make a depth-first traversal of the glycan graph. Children are explored in descending bond-order.

        This is the default traversal method for all |Glycan| objects. :meth:`~.dfs` is an alias of this method.
        Both names can be used to specify this strategy to :meth:`~._get_traversal_method`.

        When selecting an iteration strategy, this strategy is specified as "dfs".

        Parameters
        ----------
        from_node: None or Monosaccharide
            If `from_node` is |None|, then traversal starts from the root node. Otherwise it begins
            from the given node.
        apply_fn: function
            A function applied to each node on arrival. If this function returns a non-None value,
            the result is yielded from the generator, otherwise it is ignored. Defaults to :func:`.identity`
        visited: set or None
            A :class:`set` of node ID values to ignore. If |None|, defaults to the empty `set`

        Yields
        ------
        Return Value of `apply_fn`, by default |Monosaccharide|

        See also
        --------
        Glycan.breadth_first_traversal
        '''
        node_stack = deque([self.root if from_node is None else from_node])
        visited = set() if visited is None else visited
        while len(node_stack) > 0:
            node = node_stack.pop()
            visited.add(node.id)
            if apply_fn is identity:
                yield node
            else:
                res = apply_fn(node)
                if res is not None:
                    yield res

            for link in node.links.values():
                terminal = link.parent
                if terminal.id not in visited:
                    node_stack.append(terminal)
                terminal = link.child
                if terminal.id not in visited:
                    node_stack.append(terminal)

    traversal_methods['dfs'] = "depth_first_traversal"
    traversal_methods['depth_first_traversal'] = "depth_first_traversal"

    def breadth_first_traversal(self, from_node=None, apply_fn=identity, visited=None):
        '''
        Make a breadth-first traversal of the glycan graph. Children are explored in descending bond-order.

        When selecting an iteration strategy, this strategy is specified as "bfs".

        Parameters
        ----------
        from_node: None or Monosaccharide
            If `from_node` is |None|, then traversal starts from the root node. Otherwise it begins
            from the given node.
        apply_fn: function
            A function applied to each node on arrival. If this function returns a non-None value,
            the result is yielded from the generator, otherwise it is ignored. Defaults to :func:`.identity`
        visited: set or None
            A :class:`set` of node ID values to ignore. If |None|, defaults to the empty `set`

        Yields
        ------
        Return Value of `apply_fn`, by default |Monosaccharide|

        See also
        --------
        Glycan.depth_first_traversal
        '''
        node_queue = deque([self.root if from_node is None else from_node])
        visited = set() if visited is None else visited
        while len(node_queue) > 0:
            node = node_queue.popleft()
            visited.add(node.id)

            if apply_fn is identity:
                yield node
            else:
                res = apply_fn(node)
                if res is not None:
                    yield res
            # node_queue.extend(terminal for link in node.links.values()
            #                   for terminal in link if terminal.id not in visited)
            for link in node.links.values():
                terminal = link.parent
                if terminal.id not in visited:
                    node_queue.append(terminal)
                terminal = link.child
                if terminal.id not in visited:
                    node_queue.append(terminal)

    traversal_methods['bfs'] = "breadth_first_traversal"
    traversal_methods['breadth_first_traversal'] = "breadth_first_traversal"

    def indexed_traversal(self, from_node=None, apply_fn=identity, visited=None):
        """Traverse the glycan structure along :attr:`index`.

        This is substantially faster than other traversal methods for complete traversals
        at the cost of a) requiring a call to :meth:`reindex` to populate :attr:`index`
        if it has not been called, and b) is not automatically updated if the structure is
        modified.

        When selecting an iteration strategy, this strategy is specified as "index".

        Parameters
        ----------
        from_node: None or Monosaccharide
            If `from_node` is |None|, then traversal starts from the root node. Otherwise it begins
            from the given node.
        apply_fn: function
            A function applied to each node on arrival. If this function returns a non-None value,
            the result is yielded from the generator, otherwise it is ignored. Defaults to :func:`.identity`
        visited: set or None
            A :class:`set` of node ID values to ignore. If |None|, defaults to the empty `set`

        Yields
        ------
        Return Value of `apply_fn`, by default |Monosaccharide|

        See Also
        --------
        reindex
        """
        if not self.index:
            self._build_node_index()
        if from_node is None and apply_fn is identity:
            for node in self.index:
                yield node
        else:
            i = 0
            n = len(self.index)
            if from_node is not None:
                while i < n:
                    node = self.index[i]
                    if node == from_node:
                        break
                    i += 1
            while i < n:
                node = self.index[i]
                if apply_fn is identity:
                    yield node
                else:
                    value = apply_fn(node)
                    if value is not None:
                        yield value

                i += 1

    traversal_methods['index'] = "indexed_traversal"

    def _get_traversal_method(self, method):
        """An internal helper method used to resolve traversal
        methods by name or alias.

        Parameters
        ----------
        method : :class:`str` or :class:`Callable`
            If a :class:`str`, it is looked up in the class-level
            :attr:`traversal_methods` dictionary and the name of
            the appropriate method is retrieved with :func:`getattr`
            and returned.

            If a :class:`Callable`, the function's first parameter is
            bound to `self` and returned.

        Returns
        -------
        :class:`Callable`
        """
        if method == 'dfs':
            return self.depth_first_traversal
        elif method == 'bfs':
            return self.breadth_first_traversal
        elif isinstance(method, Callable):
            return partial(method, self)
        try:
            traversal = self.traversal_methods[method]
        except KeyError:
            raise KeyError("Unknown traversal method: {}".format(method))
        traversal = getattr(self, traversal)
        return traversal

    def __iter__(self):
        return self.depth_first_traversal()

    def iternodes(self, from_node=None, apply_fn=identity, method='dfs', visited=None):
        '''
        Generic iterator over nodes dispatching to a strategy given by `method`, defaulting
        to :meth:`~.depth_first_traversal`.

        Parameters
        ----------
        from_node: None or Monosaccharide
            If `from_node` is |None|, then traversal starts from the root node. Otherwise it begins
            from the given node.
        apply_fn: function
            A function applied to each node on arrival. If this function returns a non-None value,
            the result is yielded from the generator, otherwise it is ignored. Defaults to :func:`.identity`
        method: str or `function`
            Traversal method to use. See :meth:`._get_traversal_method`
        visited: set or None
            A :class:`set` of node ID values to ignore. If |None|, defaults to the empty `set`

        Yields
        ------
        Return Value of `apply_fn`, by default Monosaccharide

        See also
        --------
        depth_first_traversal
        breadth_first_traversal
        _get_traversal_method
        '''
        traversal = self._get_traversal_method(method)
        return traversal(
            from_node=from_node, apply_fn=apply_fn, visited=visited)

    def iterlinks(self, apply_fn=identity, substituents=False, method='dfs', visited=None):
        '''
        Iterates over all |Link| objects in |Glycan|.

        Parameters
        ----------
        substituents: bool
            If `substituents` is |True|, then include the |Link| objects in
            :attr:`substituent_links` on each |Monosaccharide|
        method: str or function
            The traversal method controlling the order of the nodes visited
        visited: None or set
            The collection of id values to ignore when traversing

        Yields
        ------
        Link
        '''
        traversal = self._get_traversal_method(method)
        links_visited = set()

        def links(obj):
            if substituents:
                for pos, link in obj.substituent_links.items():
                    res = apply_fn((pos, link))
                    if res:
                        yield res
            for pos, link in obj.links.items():
                if link.id in links_visited:
                    continue
                links_visited.add(link.id)
                res = apply_fn((pos, link))
                if res:
                    yield res

        return itertools.chain.from_iterable(
            traversal(apply_fn=links, visited=visited))

    def canonicalize(self, canonicalizer=None, **kwargs):
        """Canonicalize this glycan, sorting the order in which its links from the
        same monosaccharide are traversed.

        This currently uses the the :title-reference:`GlycoCT` canonicalization
        algorithm.

        Parameters
        ----------
        canonicalizer : subclass of :class:`~.CanonicalizerBase`, optional
            The canonicalization algorithm to use
        **kwargs
            Forwarded to the canonicalizer

        Returns
        -------
        :class:`~.Glycan`
            This glycan, reordered in place.
        """
        from glypy.algorithms.canonicalize import canonicalize
        return canonicalize(self, canonicalizer=canonicalizer, **kwargs)

    def leaves(self, bidirectional=False, method='dfs', visited=None):
        '''
        Iterates over all |Monosaccharide| objects in |Glycan|, yielding only those
        that have no child nodes.

        Parameters
        ----------
        bidirectional: bool
            If `bidirectional` is |True|, then only |Monosaccharide| objects
            with only one entry in :attr:`links`.
        method: str or function
            The traversal method controlling the order of the nodes visited
        visited: None or set
            The collection of id values to ignore when traversing

        Yields
        ------
        :class:`~.Monosaccharide`
        '''
        traversal = self._get_traversal_method(method)
        if bidirectional:
            def is_leaf(obj):
                if len(obj.links) == 1:
                    yield obj
        else:
            def is_leaf(obj):
                if len(list(obj.children())) == 0:
                    yield obj

        return itertools.chain.from_iterable(
            traversal(apply_fn=is_leaf, visited=visited))

    def ambiguous_links(self):
        """Locate all links which are :class:`.AmbiguousLink` objects

        Returns
        -------
        list
            list of ambiguous links
        """
        ambiguous_links = []

        # find all ambiguous links
        for i, link in self.iterlinks():
            if link.is_ambiguous():
                ambiguous_links.append(link)
        return ambiguous_links

    def iterconfiguration(self):
        '''Iterate over all valid configurations of ambiguous linkages.

        During calculation, the :class:`~.AmbiguousLink` objects may be mutated, but
        by the time a new configuration is yielded all changes should be reversed. If an
        error occurs during configuration adjustment, it may not be possible to restore the
        object to its original state.

        Yields
        ------
        :class:`tuple` of (:class:`~.AmbiguousLink`, :class:`~.Monosaccharide`,
                           :class:`~.Monosaccharide`, :class:`int`, :class:`int`)
            The ambiguous link, the parent chosen, the child chosen, the parent linkage site chose, and the child
            linkage site chosen

        Examples
        --------
        >>> from glypy.io import glyspace
        >>> structure_record = glyspace.get("G81339YK")
        >>> structure = structure_record.structure_
        >>> configurations = []
        >>> for config_list in structure.iterconfiguration():
        ...     instance = structure.clone()
        ...     for link, conf in config_list:
        ...         link = instance.get_link(link.id)
        ...         parent = instance.get(conf[0].id)
        ...         child = instance.get(conf[1].id)
        ...         link.reconfigure(parent, child, conf[2], conf[3])
        ...     configurations.append(instance)
        >>> len(configurations)
        4

        See Also
        --------
        :meth:`~.AmbiguousLink.iterconfiguration`, :meth:`~.AmbiguousLink.reconfigure`

        '''
        ambiguous_links = self.ambiguous_links()

        combos = itertools.product(*[
            link.iterconfiguration() for link in ambiguous_links])

        for configs in combos:
            n_configs = len(configs)
            # count the number of ambiguous links with unknown linkages
            has_unknowns = []
            for i, conf in enumerate(configs):
                if conf.parent_position == UnknownPosition or conf.child_position == UnknownPosition:
                    has_unknowns.append(i)
            if not has_unknowns:
                if len(set([(c.parent, c.parent_position) for c in configs])) < n_configs:
                    continue
                yield tuple(zip(ambiguous_links, configs))
            else:
                # mask the ambiguous links so the possible attachment sites can
                # be calculated
                for link in ambiguous_links:
                    link.break_link(refund=True)

                parent_map = {
                    parent.id: MonosaccharideOccupancy.build(parent)
                    for parent, _, _, _ in configs
                }

                valid = True

                # place all concrete links
                for parent, child, parent_site, child_site in configs:
                    parent_occupancy = parent_map[parent.id]
                    if parent_site == UnknownPosition:
                        continue
                    if parent_site not in parent_occupancy.open_sites:
                        valid = False
                        break
                    # can't add more concrete links if the open sites are allocated
                    # for unknown localizations
                    elif len(parent_occupancy.open_sites) - parent_occupancy.unknown_sites <= 0:
                        valid = False
                        break
                    else:
                        # allocate the requested location so no new connections
                        # can be made there
                        ix = parent_occupancy.open_sites.index(parent_site)
                        parent_occupancy.open_sites.pop(ix)
                        parent_occupancy.occupied_sites.append(parent_site)
                if valid:
                    # place all unknown links
                    for parent, child, parent_site, child_site in configs:
                        parent_occupancy = parent_map[parent.id]
                        if parent_site != UnknownPosition:
                            continue
                        # can't add more unknown links if the open sites are allocated
                        # for unknown localizations
                        if len(parent_occupancy.open_sites) - parent_occupancy.unknown_sites <= 0:
                            valid = False
                            break
                        else:
                            parent_occupancy.unknown_sites += 1

                # re-attach ambiguous links so the state of the glycan returns to
                # its original state
                for link in ambiguous_links:
                    link.apply()

                # if configuration is valid, yield it
                if valid:
                    yield tuple(zip(ambiguous_links, configs))

    def has_undefined_linkages(self):
        """Check if this structure has undefined or ambiguous connectivity
        between its nodes.

        Returns
        -------
        bool:
            If any of its links are :class:`~.AmbiguousLink` instances, or
            have unknown positions (-1).
        """
        ambiguous_links = bool(self.ambiguous_links())
        if ambiguous_links:
            return ambiguous_links
        for node in self:
            if node.has_undefined_linkages():
                return True
        return False

    def label_branches(self):
        '''
        Labels each branch point with an alphabetical symbol. Also computes and stores
        each branch's length and stores it in :attr:`branch_lengths`. Sets :attr:`branch_lengths`
        of `self` and :attr:`Link.label` for each link attached to `self`. Also populates
        :attr:`branch_parent_map`.

        Branch symbols are increasing alphabetical characters. The root branch is denoted '-', though
        glycans having an :attr:`root` with multiple children will not have any actual branches with
        that label.

        :attr:`Link.label` updates use the current branch symbol, and the index of that link along
        that branch.

        .. note:: Labeling always uses a depth-first traversal of nodes.

        '''
        last_branch_label = MAIN_BRANCH_SYM
        self.branch_lengths = defaultdict(int)
        branch_parent_map = {}

        def parent_link_symbol(node):
            try:
                label = node.links[node.parents()[0][0]][0].label
                if label is None:
                    return MAIN_BRANCH_SYM
                else:
                    return label[0]
            except IndexError:
                return MAIN_BRANCH_SYM

        for node in self.depth_first_traversal():
            links = []
            for link in node.links.values():
                if link.is_child(node):
                    continue
                links.append(link)

            if len(links) == 1:
                label_key = parent_link_symbol(node)
                self.branch_lengths[label_key] += 1
                label = "{}{}".format(
                    label_key, self.branch_lengths[label_key])
                links[0].label = label
            else:
                last_label_key = label_key = parent_link_symbol(node)
                count = self.branch_lengths[last_label_key]
                for link in links:
                    last_branch_label = chrinc(
                        last_branch_label) if last_branch_label != MAIN_BRANCH_SYM else 'a'
                    new_label_key = last_branch_label
                    branch_parent_map[new_label_key] = last_label_key
                    self.branch_lengths[new_label_key] = count + 1
                    label = "{}{}".format(
                        new_label_key, self.branch_lengths[new_label_key])
                    link.label = label

        # Update parent branch lengths
        longest = self.branch_lengths[MAIN_BRANCH_SYM]
        for branch in sorted(list(self.branch_lengths.keys()), reverse=True):
            if branch == MAIN_BRANCH_SYM:
                continue
            length = self.branch_lengths[branch]
            longest = max(longest, length)
            parent = branch_parent_map[branch]
            self.branch_lengths[parent] = max(length, self.branch_lengths[parent])
        self.branch_parent_map = branch_parent_map
        self.branch_lengths["-"] = longest

    def count_branches(self):
        '''
        Count the number of branches in the Glycan tree

        Returns
        -------
        int
        '''
        count = 0
        for node in self:
            if len(node.links) > 2:
                count += 2 if count == 0 else 1
        return count

    def order(self, deep=False):
        '''
        The number of nodes in the graph. :meth:`__len__` is an alias of this

        Returns
        -------
        int
        '''
        count = 0
        for node in self.depth_first_traversal():
            count += 1
        return count

    def __len__(self):
        return self.order()

    @classmethod
    def register_serializer(cls, name, method):
        """Add `method` as `name` to the set of serializers to pick from
        in :meth:`serialize`

        Parameters
        ----------
        name : str
            The name of the serializer
        method : Callable
            A callable object that when called with a :class:`Glycan` returns a :class:`str`
        """
        cls._serializers[name] = method

    @classmethod
    def available_serializers(cls):
        """Get the list of available serialization formats

        Returns
        -------
        :class:`list` of :class:`str`
        """
        return list(cls._serializers.keys())

    def serialize(self, name='glycoct'):
        """Convert the structure to text.

        The serialization format is given by a :meth:`available_serializers`.

        Parameters
        ----------
        name : str, optional
            The name of the serialization format (the default is 'glycoct')

        Returns
        -------
        str
        """
        return self._serializers[name](self)

    def __repr__(self):
        """An alias for :meth:`serialize`.

        Returns
        -------
        str
        """
        return self.serialize()

    def mass(self, average=False, charge=0, mass_data=None, method='dfs'):
        '''
        Calculates the total mass of the intact graph by querying each
        node for its mass.

        Parameters
        ----------
        average: bool
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is `charge`
        mass_data: dict
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.


        Returns
        -------
        float

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        if charge == 0:
            return sum(
                node.mass(average=average, charge=0, mass_data=mass_data) for node in self.iternodes(method=method))
        else:
            return self.total_composition().calc_mass(average=average, charge=charge, mass_data=mass_data)

    def total_composition(self, method='dfs'):
        '''
        Computes the sum of the composition of all |Monosaccharide| objects in ``self``

        Returns
        -------
        :class:`~glypy.composition.Composition`
        '''

        return sum((node.total_composition() for node in self.iternodes(method=method)), Composition())

    def clone(self, index_method='dfs', visited=None, cls=None):
        '''
        Create a copy of `self`, indexed using `index_method`, a *traversal method*  or |None|.

        Parameters
        ----------
        index_method: :class:`str`
            The indexing method to use when constructing the index of the copied
            structure
        visited: :class:`set`, optional
            A set of nodes to omit traversing through during the copying processing
        cls: :class:`type`
            A subclass of :class:`Glycan`, defaulting to :attr:`__class__`

        Returns
        -------
        :class:`~glypy.structure.glycan.Glycan`
        '''
        if cls is None:
            cls = self.__class__
        clone_root = graph_clone(self.root, visited=visited)
        duplicate = cls(root=clone_root, index_method=index_method)

        return duplicate

    def exact_ordering_equality(self, other):
        '''
        Two glycans are considered equal if they are identically ordered nodes.

        Parameters
        ----------
        self, other: :class:`~glypy.structure.glycan.Glycan`

        Returns
        -------
        bool

        See also
        --------
        :meth:`glypy.structure.Monosaccharide.exact_ordering_equality`
        :term:`Exact Matching`
        '''
        if other is None:
            return False
        elif not isinstance(other, Glycan):
            return False
        return self.root.exact_ordering_equality(other.root)

    def topological_equality(self, other):
        '''
        Two glycans are considered equal if they are topologically equal.

        Parameters
        ----------
        self: :class:`Glycan`
        other: :class:`Glycan`

        Returns
        -------
        bool

        See also
        --------
        :meth:`glypy.structure.Monosaccharide.topological_equality`
        :term:`Topological Matching`
        '''
        if other is None:
            return False
        elif not isinstance(other, Glycan):
            return False
        return self.root.topological_equality(other.root)

    def __eq__(self, other):
        """Test for exact ordering equality

        Parameters
        ----------
        other : :class:`Glycan`

        Returns
        -------
        :class:`bool`

        See Also
        --------
        :meth:`exact_ordering_equality`
        :term:`Exact Matching`
        """
        return self.exact_ordering_equality(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        """Hashes the structure from the GlycoCT text representation

        See Also
        --------
        :meth:`serialize`
        """
        return hash(self.serialize("glycoct"))

    def substructures(self, max_cleavages=1, min_cleavages=1, inplace=False):
        '''
        Generate disjoint subtrees from this glycan by removing one or
        more monosaccharide-monosaccharide bond.


        Parameters
        ----------
        max_cleavages: |int|
            The maximum number of bonds to break per substructure
        min_cleavages: |int|
            The minimum number of bonds to break per substructure
        min_size: |int|
            The minimum number of monosaccharides per substructure

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        structure = self
        if not inplace:
            structure = self.clone()
        for frag in structure.break_links_subtrees(
                max_cleavages):
            yield frag

    def name_fragment(self, fragment):
        '''
        Attempt to assign a full name to a fragment based on the branch and position relative to
        the reducing end along side A/B/C/X/Y/Z, according to :title-reference:`Domon and Costello`

        The formal grammar for fragment names in Backus-Naur Form:

        .. code:: xml

            <full-name>                ::= <fragment-name>|<fragment-name-list>
            <fragment-name>            ::= <glycosidic-fragment-name>|<crossring-fragment-name>
            <fragment-name-list>       ::= <fragment-name>"-"<fragment-name-list>|<fragment-name>
            <glycosidic-fragment-name> ::= <branch-identifier><fragment-type><index>
            <crossring-fragment-name>  ::= <ring-coordinates><fragment-type><branch-identifier><index>
            <fragment-type>            ::= "A" | "B" | "C" | "X" | "Y" | "Z"
            <ring-coordinate>          ::= <integer>,<integer>
            <index>                    ::= <integer>
            <integer>                  ::= <digit>|<integer><digit>
            <digit>                    ::= "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
            <branch-identifier>        ::= <letter>|<letter><digit>|""
            <letter>                   ::= "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i" | "j" |
                                           "k" | "l" | "m" | "n" | "o" | "p" | "q" | "r" | "s" | "t" |
                                           "u" | "v" | "w" | "x" | "y" | "z"

        '''

        break_targets = fragment.link_ids
        crossring_targets = fragment.crossring_cleavages

        # Accumulator for name components
        name_parts = []
        # Collect cross-ring fragment names
        for crossring_id in crossring_targets:
            # Seek the link that holds the fragmented residue
            for link in self.link_index:
                if link.child.id == crossring_id:
                    ion_type = crossring_targets[crossring_id]
                    label = link.label
                    if _fragment_direction[ion_type[1]] > 0:
                        name = "{}{}".format(
                            ''.join(map(str, ion_type)), label.replace(MAIN_BRANCH_SYM, ""))
                        name_parts.append(name)
                    else:
                        label_key = label[0]
                        distance = int(label[1:])
                        inverted_distance = self.branch_lengths[
                            label_key] - (distance - 1)
                        name = "{}{}{}".format(
                            ''.join(map(str, ion_type)), label_key.replace(MAIN_BRANCH_SYM, ""), inverted_distance)
                        name_parts.append(name)

        # Collect glycocidic fragment names
        for break_id, ion_type in break_targets.items():
            ion_type = ion_type[1]
            if _fragment_direction[ion_type] > 0:
                link = self.link_index[break_id - 1]
                label = link.label
                name = "{}{}".format(
                    ion_type,
                    label.replace(
                        MAIN_BRANCH_SYM,
                        ""))
                name_parts.append(name)
            else:
                link = self.link_index[break_id - 1]
                label = link.label
                label_key = label[0]
                distance = int(label[1:])
                inverted_distance = self.branch_lengths[
                    label_key] - (distance - 1)
                name = "{}{}{}".format(
                    ion_type, label_key.replace(MAIN_BRANCH_SYM, ""), inverted_distance)
                name_parts.append(name)

        return '-'.join(sorted(name_parts))

    def break_links_subtrees(self, n_links):
        r"""Iteratively generate all subtrees from glycosidic bond cleavages, creating all
        :math:`2{L \choose n}` subtrees.

        Parameters
        ----------
        n_links : int
            Number of links to break simultaneously

        Yields
        ------
        Subtree
        """
        if len(self.link_index) == 0:
            self._build_link_index()
        links = list(self.link_index)
        for breaks in itertools.combinations(links, n_links):

            subtrees = []
            for link in breaks:
                parent, child = link.break_link(refund=True)
                parent_tree = Glycan(parent, index_method=None)
                child_tree = Glycan(child, index_method=None)
                subtrees.append(parent_tree)
                subtrees.append(child_tree)

            unique_subtrees = []
            for subtree in subtrees:
                ids = {n.id for n in subtree}
                for uids, unique in unique_subtrees:
                    if ids == uids:
                        break
                else:
                    unique_subtrees.append((ids, subtree))

            for ids, subtree in unique_subtrees:
                subtree = subtree.clone(
                    index_method=None).reroot(
                    index_method=None)
                include_nodes = {n.id for n in subtree}

                link_ids = [link.id for link in breaks
                            if link.parent.id in include_nodes or
                            link.child.id in include_nodes]

                parent_break_ids = {link.id: link.parent.id for link in breaks
                                    if link.parent.id in include_nodes}

                child_break_ids = {link.id: link.child.id for link in breaks
                                   if link.child.id in include_nodes}

                yield Subtree(subtree, include_nodes, link_ids, parent_break_ids, child_break_ids)

            for link in breaks:
                link.apply()

    def crossring_subtrees(self, n_links):
        """Generate all combinations of cross ring fragments and
        glycosidic cleavages, cleaving between 1 and `n_links`
        monosaccharides paired with `n_links` - 1 to 0 glycosidic cleavages.

        Parameters
        ----------
        n_links : int
            Total number of breaks to create, between cross ring cleavages and
            complemenatary glycosidic cleavages.

        Yields
        ------
        Subtree
        """
        if len(self.link_index) == 0:
            self._build_link_index()

        links = list(self.link_index)
        # origin_mass = self.mass()
        # Localize globals
        _str = str
        # Break at least one ring
        for i in range(1, n_links + 1):
            # Generate all combinations of i rings to break
            for link_combination in itertools.combinations(links, i):
                # Creates a list of lists of CrossRingPairs, each inner list for a
                # single Monosaccharide
                crossring_combinations = [
                    CrossRingPair.from_link(link) for link in link_combination]
                # Combinations are splatted to unwrap the outer container so that
                # the inner lists are multiplexed.
                for breaks in itertools.product(*crossring_combinations):
                    subtrees = []
                    for ring in breaks:
                        parent, child = ring.break_link()
                        parent_tree = Glycan(parent, index_method=None)
                        parent_tree._build_node_index()
                        child_tree = Glycan(child, index_method=None)
                        child_tree._build_node_index()
                        subtrees.append(parent_tree)
                        subtrees.append(child_tree)

                    unique_subtrees = []
                    for tree in subtrees:
                        ids = {n.id for n in tree.index}
                        for uids, unique in unique_subtrees:
                            if ids == uids:
                                break
                        else:
                            unique_subtrees.append((ids, tree))

                    # If tis iteration hasn't broken all n rings, there are some
                    # breaks left over to be generated in glycosidic cleavages.
                    # Generate all possible glycosidic cleavage subtrees of the
                    # generated cross ring cleavage subtrees.

                    if n_links - i > 0:
                        for ids, subtree in unique_subtrees:
                            subtree._build_link_index()
                            partitions = subtree.break_links_subtrees(n_links - i)
                            for part in partitions:
                                included_crossring = {}
                                for crossring in breaks:
                                    if crossring.id in part.include_nodes:
                                        xring_residue = part.tree.get(
                                            crossring.id)
                                        try:
                                            included_crossring[xring_residue.id] = (','.join(
                                                (_str(xring_residue.cleave_1), _str(xring_residue.cleave_2))),
                                                xring_residue.kind)
                                        except AttributeError:
                                            pass
                                            # Not a CrossRingFragment in this instance

                                part.crossring_cleavages = included_crossring
                                yield part
                    else:
                        for ids, subtree in unique_subtrees:
                            subtree = subtree.reroot(
                                index_method=None).clone(
                                index_method=None)

                            included_crossring = {}
                            include_nodes = {n.id for n in subtree.indexed_traversal()}
                            for crossring in breaks:
                                if crossring.id in include_nodes:
                                    xring_residue = subtree.get(crossring.id)
                                    try:
                                        included_crossring[xring_residue.id] = (','.join(
                                            (_str(xring_residue.cleave_1), _str(xring_residue.cleave_2))),
                                            xring_residue.kind)
                                    except AttributeError:
                                        pass
                                        # Not a CrossRingFragment in this instance
                            yield Subtree(subtree, include_nodes, {}, {}, {}, crossring_cleavages=included_crossring)
                    # Re-join the ring, retracting all links from the crossring objects
                    for ring in breaks:
                        ring.apply()
                    # Clean up any lingering links trapped in the closure created by adjacent
                    # crossring pairs that were made visible by `ring.apply()`
                    while(any(ring.is_attached() for ring in breaks)):
                        for ring in breaks:
                            ring.release()
                    for ring in breaks:
                        ring.release()
                    # assert round(self.mass(), 4) == round(origin_mass, 4)

    def fragments(self, kind="BY", max_cleavages=1, average=False, charge=0, mass_data=None,
                  traversal_method='dfs'):
        '''
        Generate carbohydrate backbone fragments from this glycan by examining the disjoint subtrees
        created by removing one or more monosaccharide-monosaccharide bond.


        Parameters
        ----------
        kind: :class:`Iterable`
            Any :class:`Iterable` of characters corresponding to A/B/C/X/Y/Z
            as published by :title-reference:`Domon and Costello`
        max_cleavages: |int|
            The maximum number of bonds to break per fragment
        average: bool, optional, defaults to `False`
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is `charge`
        mass_data: dict, optional, defaults to `None`
            If mass_data is |None|, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of `mass_data` are assumed to contain elemental mass and isotopic abundance information.

        Yields
        ------
        :class:`GlycanFragment`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        :meth:`subtrees`
        :meth:`crossring_subtrees`
        :meth:`.Subtree.to_fragments`
        '''
        seen = set()
        source = self.clone()
        for i in range(1, max_cleavages + 1):
            gen = source.break_links_subtrees(i)
            if len(set("AX") & set(kind)) > 0:
                gen = itertools.chain(
                    gen,
                    source.crossring_subtrees(i))
            for subtree in gen:
                for fragment in subtree.to_fragments(kind, average=average,
                                                     charge=charge, mass_data=mass_data,
                                                     traversal_method=traversal_method):
                    fragment.name = self.name_fragment(fragment)
                    if fragment.name in seen:
                        continue
                    else:
                        seen.add(fragment.name)
                    yield fragment

    def subtrees(self, max_cleavages=1, include_crossring=False):
        '''
        Generate subtrees from this tree by breaking `max_cleavages` bonds or rings.

        Parameters
        ----------
        max_cleavages: int
            The maximum number of bonds to break per fragment
        include_crossring: bool
            Whether to include cross ring cleavages

        Yields
        ------
        Subtree
        '''
        source = self.clone()
        for i in range(1, max_cleavages + 1):
            gen = source.break_links_subtrees(i)
            if include_crossring:
                gen = itertools.chain(
                    gen,
                    source.crossring_subtrees(i))
            for subtree in gen:
                yield subtree


class NamedGlycan(Glycan):
    def __init__(self, name=None, *args, **kwargs):
        self.name = name
        super(NamedGlycan, self).__init__(*args, **kwargs)

    def clone(self, index_method='dfs', visited=None, cls=None):
        if cls is None:
            cls = NamedGlycan
        inst = super(NamedGlycan, self).clone(index_method=index_method, visited=visited, cls=cls)
        inst.name = self.name
        return inst

    def __repr__(self):
        rep = super(NamedGlycan, self).__repr__()
        return "%s\n%s" % (self.name, rep)




cdef class CComposition(dict):
    cdef object _mass
    cdef tuple _mass_args
    cpdef CComposition clone(self)
    cpdef double calc_mass(self, int average=?, charge=?, dict mass_data=?) except -1
    cpdef _from_formula(self, str formula, dict mass_data)
    cpdef _from_dict(self, comp)
    cdef long getitem(self, str elem)
    cdef void setitem(self, str elem, long val)

    cdef CComposition add_from(self, other)
    cdef CComposition subtract_from(self, other)

    @staticmethod
    cdef CComposition _create(CComposition inst)


cdef: 
    str _atom
    str _formula
    str _isotope_string

    object isotope_pattern
    object formula_pattern

    cdef str _parse_isotope_string(str label, int* isotope_num)
    cdef str _make_isotope_string(str element_name, int isotope_num)


cpdef double calculate_mass(CComposition composition=?, str formula=?, int average=?, charge=?, mass_data=?) except -1




cdef class EnumValue(object):
    cdef:
        public object group
        public str name
        public object value
        public set names
        public object _hash

    cpdef add_name(self, basestring name, bint force=*)
    cpdef int int_value(self) except -1


cdef class IntEnumValue(EnumValue):
    cdef:
        public int _int_value


cdef class EnumMeta(type):
    cdef:
        public dict name_map
        public dict value_map

    cpdef EnumValue name(self, v)
    cpdef translate(self, k)

    cdef EnumValue get(self, k)
    cdef int put(self, name, value) except -1


cdef str QMARK


'''
An implementation of a subset of the following symbolic nomenclature:
    http://www.ncbi.nlm.nih.gov/books/NBK310273/table/symbolnomenclature.T.monosaccharide_symb/?report=objectonly
'''

import logging
from collections import Counter
from functools import partial

import numpy as np
from matplotlib.path import Path
from matplotlib.textpath import TextPath
import matplotlib.patches as patches
import matplotlib
from matplotlib.colors import rgb2hex
from matplotlib.transforms import Affine2D

import glypy
from glypy.structure import Modification, Stem
from glypy.utils.enum import Enum
from glypy.io.nomenclature import identity

from .common import MonosaccharidePatch
from .symbolic_nomenclature import SymbolicNomenclatureBase

logger = logging.getLogger(__name__)

default_line_color = 'black'
default_scale_factor = 1.0
zorder = 2
line_weight = 0.5


NeuAc = glypy.monosaccharides.NeuAc
NeuGc = glypy.monosaccharides.NeuGc


class CFGNomenclature(SymbolicNomenclatureBase):

    class ResidueColor(Enum):
        gal = rgb2hex((255 / 255., 255 / 255., 0 / 255.))  # yellow
        lyx = rgb2hex((255 / 255., 255 / 255., 0 / 255.))  # yellow
        glc = rgb2hex((0 / 255., 0 / 255., 255 / 255.))  # blue
        man = rgb2hex((0, 200 / 255., 50 / 255.))  # green
        fuc = rgb2hex((255 / 255., 0 / 255., 0 / 255.))  # red
        xyl = rgb2hex((250 / 255., 234 / 255., 213 / 255.))  # orange
        neuac = rgb2hex((200 / 255., 0 / 255., 200 / 255.))  # purple
        neugc = rgb2hex((233 / 255., 255 / 255., 255 / 255.))  # light blue
        kdn = rgb2hex((0, 200 / 255., 50 / 255.))  # green
        glca = rgb2hex((0 / 255., 0 / 255., 255 / 255.))  # blue
        idoa = rgb2hex((150 / 255., 100 / 255., 50 / 255.))  # tan
        gala = rgb2hex((255 / 255., 255 / 255., 0 / 255.))  # yellow
        mana = rgb2hex((0, 200 / 255., 50 / 255.))  # green
        generic = 'white'

    class ResidueShape(Enum):
        circle = 1
        square = 2
        bisected_square = 3
        triangle = 4
        star = 5
        diamond = 6
        top_bisected_diamond = 7
        left_bisected_diamond = 8
        right_bisected_diamond = 9
        bottom_bisected_diamond = 10
        generic = 11

    def residue_color(self, monosaccharide):
        '''
        Determine which color to use to represent `monosaccharide` under the CFG
        symbol nomenclature.

        Parameters
        ----------
        monosaccharide: |Monosaccharide|
            The residue to be rendered

        Returns
        -------
        ResidueColor.EnumValue

        '''
        if any(mod == Modification.a for p, mod in monosaccharide.modifications.items()):
            return self.resolve_acid_color(monosaccharide)
        if "hex" in [monosaccharide.superclass]:
            if any(mod == Modification.d for p, mod in monosaccharide.modifications.items()) and\
                    monosaccharide.stem == (Stem.gal,):
                return self.ResidueColor.fuc
        try:
            return self.ResidueColor[monosaccharide.stem[0]]
        except KeyError:
            return self.ResidueColor.generic

    def resolve_acid_color(self, monosaccharide):
        '''
        Resolve the special case in :func:`residue_color` for acidic residues
        '''
        if ('gro' in monosaccharide.stem) and ('gal' in monosaccharide.stem):
            if any(sub.name == 'n_acetyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueColor.neuac
            elif any(sub.name == 'n_glycolyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueColor.neugc
            else:
                return self.ResidueColor.kdn
        elif 'glc' in monosaccharide.stem:
            return self.ResidueColor.glca
        elif 'gal' in monosaccharide.stem:
            return self.ResidueColor.gala
        elif 'man' in monosaccharide.stem:
            return self.ResidueColor.mana
        elif 'ido' in monosaccharide.stem:
            return self.ResidueColor.idoa

    def residue_shape(self, monosaccharide):
        '''
        Determine which shape to use to represent `monosaccharide` under the CFG
        symbol nomenclature.

        Parameters
        ----------
        monosaccharide: |Monosaccharide|
            The residue to be rendered

        Returns
        -------
        ResidueShape.EnumValue

        '''
        if any(mod == Modification.a for p, mod in monosaccharide.modifications.items()):
            return self.resolve_acid_shape(monosaccharide)
        if "hex" in [monosaccharide.superclass]:
            if any(sub.name == 'n_acetyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueShape.square
            elif any(sub.name == 'amino' or sub.name.startswith("n_") for p, sub in monosaccharide.substituents()):
                return self.ResidueShape.bisected_square
            elif any(mod == Modification.d for p, mod in monosaccharide.modifications.items()):
                return self.ResidueShape.triangle

            return self.ResidueShape.circle
        elif "pen" in [monosaccharide.superclass]:
            if 'xyl' in monosaccharide.stem or 'lyx' in monosaccharide.stem:
                return self.ResidueShape.star
        return self.ResidueShape.generic

    def resolve_acid_shape(self, monosaccharide):
        '''
        Resolve the special case in :func:`residue_shape` for acidic residues
        '''
        if ('gro' in monosaccharide.stem) and ('gal' in monosaccharide.stem):
            if any(sub.name == 'n_acetyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueShape.diamond
            elif any(sub.name == 'n_glycolyl' for p, sub in monosaccharide.substituents()):
                return self.ResidueShape.diamond
            else:
                return self.ResidueShape.diamond
        elif 'glc' in monosaccharide.stem:
            return self.ResidueShape.top_bisected_diamond
        elif 'gal' in monosaccharide.stem:
            return self.ResidueShape.left_bisected_diamond
        elif 'man' in monosaccharide.stem:
            return self.ResidueShape.right_bisected_diamond
        elif 'ido' in monosaccharide.stem:
            return self.ResidueShape.bottom_bisected_diamond
        else:
            return self.ResidueShape.bottom_bisected_diamond

    def draw_circle(self, ax, x, y, color, scale=0.1):
        path = Path(Path.unit_circle().vertices * scale, Path.unit_circle().codes)
        trans = Affine2D().translate(x, y)
        t_path = path.transformed(trans)
        patch = patches.PathPatch(
            t_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a,))
        return ma

    def draw_square(self, ax, x, y, color, scale=0.1):
        square_verts = np.array([
            (0.5, 0.5),
            (0.5, -0.5),
            (-0.5, -0.5),
            (-0.5, 0.5),
            (0.5, 0.5),
            (0., 0.),
        ]) * 2
        square_codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
        ]
        path = Path(square_verts * scale, square_codes)
        trans = Affine2D().translate(x, y)
        t_path = path.transformed(trans)
        patch = patches.PathPatch(
            t_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a,))
        return ma

    def draw_triangle(self, ax, x, y, color, scale=0.1):
        path = Path(Path.unit_regular_polygon(3).vertices * scale, Path.unit_regular_polygon(3).codes)
        trans = Affine2D().translate(
            x, y).rotate_deg_around(x, y, 0)
        t_path = path.transformed(trans)
        patch = patches.PathPatch(
            t_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a,))
        return ma

    def draw_bisected_square(self, ax, x, y, color, scale=0.1):
        lower_verts = (np.array([
            (0., 0.),
            (1.0, 0),
            (0, 1.0),
            (0, 0),
            (0., 0.),
        ]) - 0.5) / 5

        upper_verts = (np.array([
            (1., 1.),
            (1.0, 0),
            (0, 1.0),
            (1, 1),
            (0., 0.),
        ]) - 0.5) / 5

        codes = [Path.MOVETO,
                 Path.LINETO,
                 Path.LINETO,
                 Path.LINETO,
                 Path.CLOSEPOLY,
                 ]

        lower_path = Path(lower_verts, codes).transformed(
            Affine2D().translate(x, y))
        upper_path = Path(upper_verts, codes).transformed(
            Affine2D().translate(x, y))

        patch = patches.PathPatch(
            lower_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        patch = patches.PathPatch(
            upper_path, facecolor="white", lw=line_weight, zorder=2)
        b = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a, b))
        return ma

    def draw_diamond(self, ax, x, y, color, scale=0.1):
        path = Path(Path.unit_regular_polygon(4).vertices * scale, Path.unit_regular_polygon(4).codes)
        trans = Affine2D().translate(x, y)
        t_path = path.transformed(trans)
        patch = patches.PathPatch(
            t_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = (ax.add_patch(patch),)
        ma = MonosaccharidePatch(saccharide_shape=(a,))
        return ma

    def draw_right_bisected_diamond(self, ax, x, y, color, scale=0.1):
        return self.draw_horizontal_bisected_diamond(ax, x, y, color, scale, 'left')

    def draw_left_bisected_diamond(self, ax, x, y, color, scale=0.1):
        return self.draw_horizontal_bisected_diamond(ax, x, y, color, scale, 'right')

    def draw_top_bisected_diamond(self, ax, x, y, color, scale=0.1):
        return self.draw_vertical_bisected_diamond(ax, x, y, color, scale, 'top')

    def draw_bottom_bisected_diamond(self, ax, x, y, color, scale=0.1):
        return self.draw_vertical_bisected_diamond(ax, x, y, color, scale, 'bottom')

    def draw_vertical_bisected_diamond(self, ax, x, y, color, scale=0.1, side=None):
        lower_verts = (np.array([
            (0., 0.),
            (1.0, 0),
            (0, 1.0),
            (0, 0),
            (0., 0.),
        ]) - 0.5) / 5

        upper_verts = (np.array([
            (1., 1.),
            (1.0, 0),
            (0, 1.0),
            (1, 1),
            (0., 0.),
        ]) - 0.5) / 5

        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
        ]

        lower_path = Path(lower_verts, codes).transformed(
            Affine2D().translate(x, y).rotate_deg_around(x, y, 45))
        upper_path = Path(upper_verts, codes).transformed(
            Affine2D().translate(x, y).rotate_deg_around(x, y, 45))

        try:
            color = color.value
        except AttributeError:
            color = 'white'

        if side == 'top':
            top_color = color
            bottom_color = 'white'
        elif side == 'bottom':
            top_color = 'white'
            bottom_color = color
        patch = patches.PathPatch(
            lower_path, facecolor=bottom_color, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        patch = patches.PathPatch(
            upper_path, facecolor=top_color, lw=line_weight, zorder=2)
        b = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a, b))
        return ma
        # return a, b

    def draw_horizontal_bisected_diamond(self, ax, x, y, color, scale=0.1, side=None):
        left_verts = (np.array([
            (0., 0.),
            (1.0, 0),
            (0, 1.0),
            (0, 0),
            (0., 0.),
        ]) - 0.5) / 5

        right_verts = (np.array([
            (1., 1.),
            (1.0, 0),
            (0, 1.0),
            (1, 1),
            (0., 0.),
        ]) - 0.5) / 5

        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY,
        ]

        try:
            color = color.value
        except AttributeError:
            color = 'white'

        if side == 'left':
            left_color = color
            right_color = 'white'
        elif side == 'right':
            left_color = 'white'
            right_color = color
        left_path = Path(left_verts, codes).transformed(
            Affine2D().translate(x, y).rotate_deg_around(x, y, -45))
        right_path = Path(right_verts, codes).transformed(
            Affine2D().translate(x, y).rotate_deg_around(x, y, -45))

        patch = patches.PathPatch(
            left_path, facecolor=left_color, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        patch = patches.PathPatch(
            right_path, facecolor=right_color, lw=line_weight, zorder=2)
        b = ax.add_patch(patch)
        ma = MonosaccharidePatch(
            saccharide_shape=(a, b), saccharide_label=None)
        return ma
        # return a, b

    def draw_star(self, ax, x, y, color, scale=0.1):
        path = Path(Path.unit_regular_star(5, 0.45).vertices * scale, Path.unit_regular_star(5, 0.45).codes)
        trans = Affine2D().translate(x, y)
        t_path = path.transformed(trans)
        patch = patches.PathPatch(
            t_path, facecolor=color.value, lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a,))
        return ma

    def draw_generic(self, ax, x, y, name, n_points=6, scale=0.1):
        unit_polygon = Path.unit_regular_polygon(n_points)
        path = Path(unit_polygon.vertices * scale, unit_polygon.codes)
        trans = Affine2D().translate(x, y)
        t_path = path.transformed(trans)
        name = TextPath((x - (0.35 * scale), y), s=name, size=2 * scale * .25)
        patch = patches.PathPatch(
            t_path, facecolor="white", lw=line_weight, zorder=2)
        a = ax.add_patch(patch)
        patch = patches.PathPatch(name, lw=line_weight, zorder=2)
        s = ax.add_patch(patch)
        ma = MonosaccharidePatch(saccharide_shape=(a,), saccharide_label=(s,))
        return ma

    def get_drawer(self, shape):
        shape_name = shape.name
        drawer = getattr(self, 'draw_%s' % (shape_name,), None)
        return drawer

    def resolve_generic_name(self, monosaccharide):
        try:
            abbrev = identity.identify(monosaccharide)
        except identity.IdentifyException:
            if monosaccharide.stem[0] == Stem.x:
                abbrev = monosaccharide.superclass.name.lower().capitalize()
            else:
                abbrev = ','.join(s.name.lower().capitalize()
                                  for s in monosaccharide.stem)
        return abbrev

    def get_relevant_substituents(self, residue, shape=None):
        '''
        Given the shape for a residue, determine which of its substituents must
        be explicitly drawn.  Calls :func:`residue_shape` if `shape` is not
        provided.
        Parameters
        ----------
        residue: |Monosaccharide|
            The monosaccharide residue being rendered
        shape: ResidueShape or |None|
            The shape enum being used to represent `residue`. Defaults to None.
            If `shape` is |None|, it is calculated by :func:`residue_shape`.
        Returns
        -------
        |list| of |Substituent|s
        '''

        shape = self.residue_shape(residue) if shape is None else shape
        if shape != self.ResidueShape.generic:
            substituents = list(sub.name for p, sub in residue.substituents() if not sub._derivatize)
            if shape == self.ResidueShape.square:
                substituents.pop(substituents.index("n_acetyl"))
            elif shape == self.ResidueShape.bisected_square:
                try:
                    index_amino = substituents.index("amino")
                    substituents.pop(index_amino)
                except ValueError:
                    pass
                try:
                    pass
                    # index_n_sulfate = substituents.index('n_sulfate')
                    # TODO:
                    # Find a way to forward different substituents
                    #
                    # substituents.pop(index_n_sulfate)
                    # substituents.append("sulfate")
                except ValueError:
                    pass
                return self._pruned_substituents(residue, substituents)
            elif shape == self.ResidueShape.diamond:
                color = self.residue_color(residue)
                if color == self.residue_color(NeuAc):
                    substituents.pop(substituents.index("n_acetyl"))
                elif color == self.residue_color(NeuGc):
                    substituents.pop(substituents.index("n_glycolyl"))
            return self._pruned_substituents(residue, substituents)

        return list((p, sub.name) for p, sub in residue.substituents() if not sub._derivatize)

    def _pruned_substituents(self, residue, substituents):
        relevant_substituents = Counter(substituents)
        results = []
        for p, sub in residue.substituents()[::-1]:
            if relevant_substituents[sub.name] > 0:
                results.append((p, sub.name))
                relevant_substituents[sub.name] -= 1
        return results

    def get_symbol(self, monosaccharide):
        '''
        Convenience function to retrieve the shape and color of `monosaccharide`
        '''
        shp = self.residue_shape(monosaccharide)
        col = self.residue_color(monosaccharide)
        return shp, col

    def draw(self, monosaccharide, x, y, ax, tree_node=None, scale=0.1, annotation_transform=None, **kwargs):
        '''
        Renders `monosaccharide` at the given `(x, y)` coordinates on the `matplotlib.Axis`
        `ax` provided. Determines the shape to use by :func:`residue_shape` and color by
        :func:`residue_color`. The shape value is used to select the specialized draw_* function
        '''
        abbrev = None
        shape, color = self.get_symbol(monosaccharide)
        if shape == self.ResidueShape.generic:
            abbrev = self.resolve_generic_name(monosaccharide)
        drawer = self.get_drawer(shape)
        if drawer is None:
            raise Exception("Don't know how to draw {}".format(
                (shape, monosaccharide)))

        res = None
        if shape == self.ResidueShape.generic:
            res = drawer(
                ax, x, y, abbrev, n_points=monosaccharide.superclass.value or 1, scale=scale)
        else:
            res = drawer(ax, x, y, color, scale=scale)
        substituents = self.get_relevant_substituents(monosaccharide)

        # Render substituents along the bottom of the monosaccharide
        # These layouts should be moved to be defined by the DrawTreeNode
        if annotation_transform is None:
            annotation_transform = Affine2D()
        node_x, node_y = res.centroid()
        subs = self.draw_substituents(ax, substituents, node_x, node_y, annotation_transform, **kwargs)
        res.add_substituents(subs)
        return res

    def __call__(self, *args, **kwargs):
        return self.draw(*args, **kwargs)





cdef class MoleculeBase(object):

    cdef:
        public int _order

    cpdef order(self, deep=*)
    cpdef has_undefined_linkages(self)


cdef class SaccharideBase(MoleculeBase):
    pass


cdef class SaccharideCollection(SaccharideBase):
    cpdef _derivatized(self, substituent, id_base)

    cpdef _strip_derivatization(self)


cdef class SubstituentBase(MoleculeBase):
    pass


cdef class ModificationBase(MoleculeBase):
    pass



from .geometry import breadth_first_traversal, centroid, make_path, TreeLayoutBase
import numpy as np

#: clock-face coordinates assume left-horizontal growth
index_position_shift = {
    0: (-0.0, -0.0),
    2: (-0.0, 1),
    3: (-0.5, 1),
    4: (-0.8, 0),
    5: (-0.5, -1),
    6: (-0.6, -1),
    8: (-0, -1)
}
index_position_shift = {k: np.array(v[::-1], dtype=np.float64) for k, v in index_position_shift.items()}


class TopologicalTreeLayout(TreeLayoutBase):

    def __init__(self, root, **kwargs):
        super(TopologicalTreeLayout, self).__init__(root, **kwargs)
        self.linkage_index = dict()

    def before_layout(self, **kwargs):
        for node in self.traverse():
            node.mask_special_cases = False

    def after_layout(self, **kwargs):
        """Apply any cleanup after applying layout
        """
        pass

    def layout_tree(self, **kwargs):
        visited = set()
        iterations = kwargs.get("iterations", 25)
        self.layout_node(self.root, 0, 0, 0, visited)
        for i in range(int(iterations)):
            delta = self.adjust_subtrees(self.root)
            if delta == 0:
                break

    def position_relative_to_parent(self, cx, cy, index_position, scale=1.0):
        sx, sy = index_position_shift[index_position]
        sx *= scale
        sy *= scale
        return cx - sx, cy - sy

    def layout_node(self, node, cx, cy, parent_position=3, visited=None):
        if visited is None:
            visited = set()
        if node.id in visited:
            return node
        self.linkage_index[node.id] = parent_position
        node.x, node.y = self.position_relative_to_parent(cx, cy, parent_position)
        node.mask_special_cases = False
        # default positions for unknown linkages
        default_positions = ([3, 5, 4, 6, 2, 8])
        # when there is only one child node, and that node is unknown,
        # prefer a straight line instead of a slanted line.
        if len(node.tree.children()) == 1:
            default_positions = [4]
        deferred_children = []
        for child_node, pos_child in zip(node.children, sorted(node.tree.children())):
            pos = pos_child[0]
            if pos != -1:
                if pos in default_positions:
                    default_positions.remove(pos)
            else:
                deferred_children.append(child_node)
                continue
            self.layout_node(child_node, node.x, node.y, pos)
        if len(deferred_children) > len(default_positions):
            raise ValueError("Out of child positions")
        for child_node, pos in zip(deferred_children, default_positions):
            self.layout_node(child_node, node.x, node.y, pos)
        return node

    def adjust_subtrees(self, tree):
        next_layer = tree.children
        if len(next_layer) == 0:
            return 0

        total_shift = 0

        # layout later layers first in a depth-first traversal
        for node in next_layer:
            total_shift += self.adjust_subtrees(node)

        if len(next_layer) < 2:
            return total_shift

        next_layer = sorted(next_layer, key=lambda x: x.x, reverse=False)

        prior_nodes = [next_layer[0]]
        shifted = [next_layer[0]]
        for node in next_layer[1:]:
            for last in prior_nodes:
                # This creates a circle around each node and reports on overlaps
                # of nodes, but doesn't care about edges. The path extraction
                # function should try to cover edges too.
                lpath = make_path(last, 0.3)
                npath = make_path(node, 0.3)
                delta = 0
                # if the covering paths intersect
                if lpath.intersects_path(npath):
                    lcentroid = centroid(lpath)
                    ncentroid = centroid(npath)
                    lxmin_lymin, lxmax_lymax = lpath.get_extents().get_points()
                    nxmin_nymin, nxmax_nymax = npath.get_extents().get_points()
                    # the node is in a special position which should not be shifted
                    # according to examples
                    # if last.x == tree.x or self.linkage_index[last.id] == 4:
                    if self.linkage_index[node.id] == 4:
                        delta = (nxmax_nymax[0] - ncentroid[0]) / -2
                        self.shift_subtree(last, dx=delta)
                    elif ncentroid[0] > lcentroid[0]:
                        delta = (nxmin_nymin[0] - ncentroid[0]) / -2
                        self.shift_subtree(node, dx=delta)
                    elif ncentroid[0] < lcentroid[0]:
                        delta = (lxmax_lymax[0] - lcentroid[0]) / -2
                        self.shift_subtree(last, dx=delta)
                    total_shift += abs(delta)
            shifted.append(node)
            prior_nodes.append(node)
        return total_shift

    def shift_subtree(self, tree, dx=0, dy=0):
        for n in breadth_first_traversal(tree):
            n.x += dx
            n.y += dy
        return tree


def _test_draw_unit_perimeter(ax):  # pragma: no cover
    ax.text(0, 0, 'origin')
    for i in index_position_shift:
        ax.text(*square_perimeter(0, 0, i), s=i)
    return ax


def square_perimeter(cx, cy, index_position, scale=1.0):
    sx, sy = index_position_shift[index_position]
    sx *= scale
    sy *= scale
    return cx - sx, cy - sy


def layout_node(node, cx, cy, parent_position=3, visited=None):
    if visited is None:
        visited = set()
    node.x, node.y = square_perimeter(cx, cy, parent_position)
    node.mask_special_cases = False
    for child_node, pos_child in zip(node.children, sorted(node.tree.children())):
        pos = pos_child[0]
        layout_node(child_node, node.x, node.y, pos)
    return node


def layout(tree, visited=None):
    if visited is None:
        visited = set()
    layout_node(tree, 0, 0, 0, visited)
    for i in range(10):
        delta = spread(tree)
        if delta == 0:
            break
    return tree


def spread(tree):
    next_layer = tree.children
    if len(next_layer) == 0:
        return 0

    total_shift = 0

    # layout later layers first in a depth-first traversal
    for node in next_layer:
        total_shift += spread(node)

    if len(next_layer) < 2:
        return total_shift

    # layout this layer in x-ascending order
    next_layer = sorted(next_layer, key=lambda x: x.x)

    prior_nodes = [next_layer[0]]
    shifted = [next_layer[0]]
    for node in next_layer[1:]:
        for last in prior_nodes:
            lpath = make_path(last)
            npath = make_path(node)
            delta = 0
            if lpath.intersects_path(npath):
                lcentroid = centroid(lpath)
                ncentroid = centroid(npath)
                lxmin_lymin, lxmax_lymax = lpath.get_extents().get_points()
                nxmin_nymin, nxmax_nymax = npath.get_extents().get_points()
                if last.x == tree.x:
                    delta = (nxmax_nymax[0] - ncentroid[0]) / 2
                    shift_subtree(node, delta)
                elif ncentroid[0] > lcentroid[0]:
                    delta = (nxmin_nymin[0] - ncentroid[0]) / -2
                    shift_subtree(node, delta)
                elif ncentroid[0] < lcentroid[0]:
                    delta = (lxmax_lymax - lcentroid[0]) / -2
                    shift_subtree(last, delta)
                total_shift += abs(delta)
        shifted.append(node)
        prior_nodes.append(node)
    return total_shift


def shift_subtree(tree, dx):
    for n in breadth_first_traversal(tree):
        n.x += dx
    return tree



import re

from matplotlib import patches
from matplotlib import transforms as mtransforms
from matplotlib.textpath import TextPath

from glypy.plot.symbolic_nomenclature import line_to, line_along
from glypy.plot.geometry import l2_distance, centroid


def isclose(a, b):
    return abs(a - b) < 1e-4


debug = False
rot90 = mtransforms.Affine2D().rotate_deg(90)


def midpoint(a, b):
    return [(a[0] + b[0]) / 2, (a[1] + b[1]) / 2]


class BondCleavageArtist(object):

    def __init__(self, draw_tree, fragment, ax, label=True, **graphic_options):
        self.draw_tree = draw_tree
        self.fragment = fragment
        self.ax = ax
        self.artists = []
        self.patches = []
        self.label = label
        self.graphic_options = graphic_options
        self.handle_glycosidic()
        self.handle_crossring()
        self.propagate_transform()

    def handle_glycosidic(self):
        for link_id, frag_type in self.fragment.link_ids.items():
            parent, child = self.draw_tree.get_link_pair(link_id)
            if frag_type[1] in "YZ":
                reducing = True
            else:
                reducing = False
            gba = GlycosidicBondCleavageStroke(self.draw_tree, parent, child, reducing,
                                               ax=self.ax, fragment=self.fragment,
                                               label=self.label,
                                               **self.graphic_options)
            gba.draw()
            self.artists.append(gba)
            self.patches.extend(gba.patches)

    def handle_crossring(self):
        for crossring_id, frag_type in self.fragment.crossring_cleavages.items():
            node = self.draw_tree.get(crossring_id)
            if frag_type[1] == "X":
                reducing = True
            else:
                reducing = False
            cra = CrossRingBondCleavageStroke(
                self.draw_tree, node, reducing, self.ax, self.fragment,
                label=self.label,
                **self.graphic_options)
            cra.draw()
            self.artists.append(cra)
            self.patches.extend(cra.patches)

    def propagate_transform(self):
        transform = self.draw_tree.transform
        if transform is None:
            return
        transform = transform + self.ax.transData
        for patch in self.patches:
            patch.set_transform(transform)


class CleavageStrokeArtistBase(object):

    def __init__(self, draw_tree, fragment, reducing, ax, label=True, **options):
        self.draw_tree = draw_tree
        self.ax = ax
        self.reducing = reducing
        self.fragment = fragment
        self.patches = []
        self.x_coords = []
        self.y_coords = []
        self.patch_dict = draw_tree.data['patches']
        self.options = options
        self.options.setdefault("fontsize", 0.5)
        self.reducing_end_coords = (self.draw_tree.x, self.draw_tree.y)
        self.edge_size = None
        self.label = label

    def save_patch(self, key, value):
        self.patches.append(value)
        if key in self.patch_dict:
            self.patch_dict[key].append([value])
        else:
            self.patch_dict[key] = [[value]]

    def textpath(self, x, y, text, line_weight=0.5, ha='center', va='center', **kwargs):
        fs = kwargs.get("fontsize", 2) * .5
        t_path = TextPath((x, y), s=text, size=fs)
        if ha == 'center':
            cx, cy = centroid(t_path)
            tf = mtransforms.Affine2D()
            tf.translate(x - cx, 0)
            t_path = t_path.transformed(tf)
        if va == 'center':
            cx, cy = centroid(t_path)
            tf = mtransforms.Affine2D()
            tf.translate(0, y - cy)
            t_path = t_path.transformed(tf)

        patch = patches.PathPatch(t_path, facecolor="black", lw=line_weight / 20., zorder=4)
        a = self.ax.add_patch(patch)
        return a


class CrossRingBondCleavageStroke(CleavageStrokeArtistBase):

    def __init__(self, draw_tree, node, reducing, ax, fragment, label=True, **options):
        super(CrossRingBondCleavageStroke, self).__init__(
            draw_tree, fragment, reducing, ax, label, **options)
        self.node = node
        self.mode = None

    def draw(self):
        self.compute_positions()
        self.draw_main_stroke()
        self.draw_edge_stroke()

    def compute_positions(self):
        px, py = self.node.x, self.node.y
        if self.node.children:
            nx, ny = -1, -1
            for next_node in self.node.children:
                cx, cy = next_node.x, next_node.y
                if cy > ny:
                    nx, ny = cx, cy
            self.mode = 'child'
            cx, cy = nx, ny
            xcenter = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            ycenter = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            xlength = (xcenter - min(px, cx)) / 2.
            ylength = (ycenter - min(py, cy)) / 2.
            if py < cy:
                lx2, ly2 = self.node.x + xlength, self.node.y + ylength
                lx1, ly1 = self.node.x - xlength, self.node.y - ylength
            else:
                lx1, ly1 = self.node.x - xlength, self.node.y + ylength
                lx2, ly2 = self.node.x + xlength, self.node.y - ylength
            edge_size = abs(ly1 - ly2)

        elif self.node.parent:
            nx, ny = self.parent
            self.mode = 'parent'

            cx, cy = px, py
            px, py = nx, ny

            xcenter = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            ycenter = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            xlength = (xcenter - min(px, cx)) / 2.
            ylength = (ycenter - min(py, cy)) / 2.
            if py < cy:
                lx2, ly2 = xcenter + xlength, ycenter + ylength
                lx1, ly1 = xcenter - xlength, ycenter - ylength
            else:
                lx1, ly1 = xcenter - xlength, ycenter + ylength
                lx2, ly2 = xcenter + xlength, ycenter - ylength
            edge_size = abs(ly1 - ly2)

        else:
            print("Orphaned Node. Cross-ring fragment drawing may be bad")

        self.x_coords = [lx1, lx2]
        self.y_coords = [ly1, ly2]
        self.edge_size = edge_size

    def draw_main_stroke(self, **kwargs):
        graphic_options = self.options.copy()
        graphic_options.update(kwargs)

        color = graphic_options.get('color', 'red')
        lw = graphic_options.get('lw', 2)

        lx1, lx2 = self.x_coords
        ly1, ly2 = self.y_coords

        line = line_to(self.ax, lx1, ly1, lx2, ly2,
                       color=color, zorder=10, lw=lw)
        self.save_patch(self.fragment.name, line)
        line.set_gid(self.draw_tree.uuid + '-' + self.fragment.name)

    def draw_edge_stroke(self, **kwargs):
        graphic_options = self.options.copy()
        graphic_options.update(kwargs)

        color = graphic_options.get('color', 'red')
        lw = graphic_options.get('lw', 2)

        lx1, lx2 = self.x_coords
        ly1, ly2 = self.y_coords

        if self.reducing:
            line = line_to(self.ax, lx1, ly2, lx1, ly2 -
                           self.edge_size, color=color, zorder=3, lw=lw)
            line.set_gid(self.draw_tree.uuid + '-' +
                         self.fragment.name + "-direction")
            self.save_patch(self.fragment.name + "_direction", line)
        else:
            line = line_to(self.ax, lx2, ly2, lx2, ly2 +
                           self.edge_size, color=color, zorder=3, lw=lw)
            line.set_gid(self.draw_tree.uuid + '-' +
                         self.fragment.name + "-direction")
            self.save_patch(self.fragment.name + "_direction", line)


class GlycosidicBondCleavageStroke(CleavageStrokeArtistBase):

    def __init__(self, draw_tree, parent, child, reducing, ax, fragment, **options):
        super(GlycosidicBondCleavageStroke, self).__init__(
            draw_tree, fragment, reducing, ax, **options)

        self.parent = parent
        self.child = child

        self.mode = None

    def draw(self):
        self.compute_positions()
        self.configure_edge_stroke()
        self.draw_stroke()
        if self.label:
            self.draw_label()

    def compute_positions(self):
        parent, child = self.parent, self.child
        px, py = parent.x, parent.y
        cx, cy = child.x, child.y

        line_length_scale = float(self.options.get("line_length_scale", 1.0))

        if isclose(py, cy) and not isclose(px, cx):
            mode = 'offsequence'
            center = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            length = (center - min(px, cx))
            length *= line_length_scale
            lx1, ly1 = center, py + length
            lx2, ly2 = center, py - length
            edge_size = abs(ly1 - ly2) / 2.

        elif not isclose(py, cy) and isclose(px, cx):
            mode = "sequential"
            center = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            length = center - min(py, cy)
            length *= line_length_scale
            lx1, ly1 = px + length, center
            lx2, ly2 = px - length, center
            edge_size = abs(lx1 - lx2) / 2.
        else:
            mode = "branching"
            xcenter = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            ycenter = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            xlength = (xcenter - min(px, cx)) / 2.
            ylength = (ycenter - min(py, cy)) / 2.
            if py < cy:
                lx2, ly2 = xcenter + xlength, ycenter + ylength
                lx1, ly1 = xcenter - xlength, ycenter - ylength
                if (py < ly1 < ly2 < cy) and (px < lx1 < lx2 < cx):
                    lx1, lx2 = lx2, lx1
            else:
                lx1, ly1 = xcenter - xlength, ycenter + ylength
                lx2, ly2 = xcenter + xlength, ycenter - ylength
            edge_size = abs(ly1 - ly2)

        self.mode = mode
        self.edge_size = edge_size
        self.point_a = [lx1, ly1]
        self.point_b = [lx2, ly2]
        self.x_coords = [lx1, lx2]
        self.y_coords = [ly1, ly2]

    def configure_edge_stroke(self, **kwargs):
        lx1, lx2 = self.x_coords
        ly1, ly2 = self.y_coords

        if self.mode in ("sequential", "offsequence"):
            self.reducing_end_coords = (self.parent.x, self.parent.y)

        if self.reducing:
            t1 = self.point_a
            t2 = self.point_b

            if debug:
                self.ax.scatter(*t1, zorder=100, color='orange')
                self.ax.scatter(*t2, zorder=100, color='green')

            # find the terminal of the main stroke that points towards the reducing end
            anchor_pt = t1
            if l2_distance(t1, self.reducing_end_coords) > l2_distance(t2, self.reducing_end_coords):
                anchor_pt = t2

            # find the axis direction that points towards the reducing end
            best_term_pt = None
            best_distance = float('inf')
            i = 0
            for sign, idx in ((1, 0), (1, 1), (-1, 0), (-1, 1)):
                new_pt = list(anchor_pt)
                new_pt[idx] += sign * self.edge_size
                dist = l2_distance(new_pt, self.reducing_end_coords)
                i += 1
                if dist < best_distance and l2_distance(midpoint(t1, t2), new_pt) > 1e-2:
                    best_term_pt = new_pt
                    best_distance = dist

            if best_term_pt is None:
                best_term_pt = anchor_pt

            if debug:
                self.ax.scatter(*best_term_pt, zorder=100, color='purple')
        else:
            t1 = self.point_a
            t2 = self.point_b

            if debug:
                self.ax.scatter(*t1, zorder=100, color='orange')
                self.ax.scatter(*t2, zorder=100, color='green')

            # find the terminal of the main stroke that points away from the reducing end
            anchor_pt = t1
            if l2_distance(t1, self.reducing_end_coords) <= l2_distance(t2, self.reducing_end_coords):
                anchor_pt = t2

            # find the axis direction that points away from the reducing end
            best_term_pt = None
            best_distance = -float('inf')
            for sign, idx in ((1, 0), (1, 1), (-1, 0), (-1, 1)):
                new_pt = list(anchor_pt)
                new_pt[idx] += sign * self.edge_size
                dist = l2_distance(new_pt, self.reducing_end_coords)
                if dist > best_distance and l2_distance(midpoint(t1, t2), new_pt) > 1e-2:
                    best_term_pt = new_pt
                    best_distance = dist

            if best_term_pt is None:
                best_term_pt = anchor_pt

            if debug:
                self.ax.scatter(*best_term_pt, zorder=100, color='purple')

        self.anchor_point = anchor_pt
        self.terminal_point = best_term_pt
        if self.anchor_point == self.point_a:
            self.start_point = self.point_b
        else:
            self.start_point = self.point_a

    def draw_stroke(self, **kwargs):
        graphic_options = self.options.copy()
        graphic_options.update(kwargs)

        color = graphic_options.get('color', 'red')
        lw = graphic_options.get('lw', 2)
        line = line_along(
            self.ax, [self.start_point, self.anchor_point, self.terminal_point],
            color=color, zorder=3, lw=lw)
        line.set_gid(self.draw_tree.uuid + '-' +
                     self.fragment.name)
        self.save_patch(self.fragment.name + "_direction", line)

    def draw_label(self, **kwargs):
        graphic_options = self.options.copy()
        graphic_options.update(kwargs)

        color = graphic_options.get('text_color', 'black')
        pad = graphic_options.get('text_pading', 0.3)
        lw = graphic_options.get('lw', 1)
        anchor_pt = self.anchor_point
        term_pt = self.terminal_point

        # the edge is vertical
        if anchor_pt[0] == term_pt[0]:
            reference_pt = midpoint(anchor_pt, term_pt)
            right_pt = list(reference_pt)
            right_pt[0] += pad
            left_pt = list(reference_pt)
            left_pt[0] -= pad

            if l2_distance(right_pt, self.start_point) < l2_distance(left_pt, self.start_point):
                reference_pt = left_pt
            else:
                reference_pt = right_pt

            if debug:
                self.ax.scatter(*reference_pt, zorder=100, color='blue')
            patch = self.textpath(
                *reference_pt, text=self.fragment.fname, ha='center',
                color=color, line_weight=lw, fontsize=self.options.get("fontsize", 0.5))
            self.save_patch("text", patch)
            self.label_point = reference_pt
        # the edge is horizontal
        elif anchor_pt[1] == term_pt[1]:
            reference_pt = midpoint(anchor_pt, term_pt)
            right_pt = list(reference_pt)
            right_pt[1] += pad / 2.0
            left_pt = list(reference_pt)
            left_pt[1] -= pad / 2.0

            if l2_distance(right_pt, self.start_point) < l2_distance(left_pt, self.start_point):
                reference_pt = left_pt
            else:
                reference_pt = right_pt

            if debug:
                self.ax.scatter(*reference_pt, zorder=100, color='blue')
            patch = self.textpath(
                *reference_pt, text=self.fragment.fname, ha='center',
                color=color, line_weight=lw, fontsize=self.options.get("fontsize", 0.5))
            self.save_patch("text", patch)
            self.label_point = reference_pt


# Old Reference Implementation
def draw_cleavage(self, fragment=None, at=(0, 0), ax=None, scale=0.1, color='red', label=True):  # pragma: no cover
    '''
    .. warning::
        Here be magical numbers. This method is highly experimental and may not behave nicely
        when projected onto a transformed topology

    '''
    if ax is None:
        ax = self.axes
    if ax is None:
        raise ValueError("`ax` is required")
    if fragment is None:
        raise ValueError("`fragment` is required")

    break_targets = fragment.link_ids
    crossring_targets = fragment.crossring_cleavages

    def isclose(a, b):
        return abs(a - b) < 1e-4

    def save_patch(data, key, value):
        _created_patches.append(value)
        if key in data:
            data[key].append([value])
        else:
            data[key] = [[value]]

    def get_scale(trans_mat):
        ty = trans_mat[1, 0]
        return ty

    def textpath(x, y, text, line_weight=0.5, **kwargs):
        fs = kwargs.get("fontsize", 2) * scale * .5
        t_path = TextPath((x, y), s=text, size=fs)
        patch = patches.PathPatch(t_path, facecolor="black", lw=line_weight / 20., zorder=4)
        a = ax.add_patch(patch)
        return a

    if self.transform:
        trans_scale = max(get_scale(self.transform.get_matrix()) * 3e-5, 0.5)
    else:
        trans_scale = 1

    scale *= 2 * trans_scale
    print(scale, trans_scale)

    patch_dict = self.data['patches']
    _created_patches = []

    for link_break in break_targets:
        parent, child = self.get_link_pair(link_break)
        px, py = parent.coords(at)
        cx, cy = child.coords(at)

        print('p', (px, py), 'c', (cx, cy))
        branch_point = False
        length = scale
        if isclose(py, cy) and not isclose(px, cx):
            center = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            length = max((center - min(px, cx)) * trans_scale, scale)
            if length < 1:
                length = length / length * 0.25
            else:
                length = length * length * 0.5
            lx1, ly1 = center, py + length
            lx2, ly2 = center, py - length
            edge_size = max(abs(ly1 - ly2) / 4000., scale * 2)
        elif not isclose(py, cy) and isclose(px, cx):
            center = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            length = max((center - min(py, cy)) * trans_scale, scale)
            if length < 1:
                length = length / length * 0.25
            else:
                length = length * length * 0.5
            lx1, ly1 = px + length, center
            lx2, ly2 = px - length, center
            edge_size = max(abs(lx1 - lx2) / 4000., scale * 2)
        else:
            branch_point = True
            xcenter = (max(px, cx) - min(px, cx)) / 2. + min(px, cx)
            ycenter = (max(py, cy) - min(py, cy)) / 2. + min(py, cy)
            xlength = (xcenter - min(px, cx)) * trans_scale
            ylength = (ycenter - min(py, cy)) * trans_scale
            if py < cy:
                lx2, ly2 = xcenter + scale, ycenter + scale
                lx1, ly1 = xcenter - scale, ycenter - scale
            else:
                lx1, ly1 = xcenter - scale, ycenter + scale
                lx2, ly2 = xcenter + scale, ycenter - scale
            edge_size = scale * 2  # max(min(abs(lx1 - lx2) / 40000., abs(ly1 - ly2) / 40000.),scale * 2)

        print("Main Line: ", (lx1, lx2), (ly1, ly2, ly2 - ly1))

        lw = 2

        line = ax.plot((lx1, lx2), (ly1, ly2), color=color, zorder=3, lw=lw)
        save_patch(patch_dict, fragment.name, line[0])
        self.data['position'][fragment.name] = (lx1, lx2), (ly1, ly2)
        line[0].set_gid(self.uuid + '-' + fragment.name)

        print((lx1, lx2), (ly1, ly2), length, edge_size)

        if fragment.link_ids[link_break][1] in {"B", "C"}:
            label_x = (lx2, lx2 - edge_size)
            if branch_point:
                label_x = [x - 2 * scale for x in label_x]

            line = ax.plot(label_x, (ly1, ly1), color=color, zorder=3, lw=lw)
            line[0].set_gid(self.uuid + '-' + fragment.name + "-direction")
            save_patch(patch_dict, fragment.name + "_direction", line[0])
            self.data['position'][fragment.name + "_direction"] = label_x, (ly1, ly1)
            if label:
                text = textpath(label_x[0] - .4, ly1 + 0.05, fragment.fname)
                save_patch(patch_dict, fragment.name + "_text", text)
                self.data['position'][fragment.name + "_text"] = label_x[0] - .4, ly1 + 0.05
        else:
            line = ax.plot((lx2, lx2 + edge_size), (ly2, ly2), color=color, zorder=3, lw=lw)
            line[0].set_gid(self.uuid + '-' + fragment.name + "-direction")
            save_patch(patch_dict, fragment.name + "_direction", line[0])
            self.data['position'][fragment.name + "_direction"] = (lx2, lx2 + scale), (ly2, ly2)

            if label:
                lx2 += (0.05 * scale)
                ly2 -= (0.15 * scale)
                save_patch(patch_dict, fragment.name + "_text", textpath(lx2 + 0.05, ly2 - 0.15,
                                                                         fragment.fname))
                self.data['position'][fragment.name + "_text"] = lx2 + 0.05, ly2 - 0.15

    for crossring in crossring_targets:
        target = self.get(crossring)
        cx, cy = target.coords(at)
        line = ax.plot((cx - scale, cx + scale), (cy + scale, cy - scale), color=color, zorder=3)
        patch_dict[fragment.name] = line[0]
        self.data['position'][fragment.name] = (cx - scale, cx + scale), (cy + scale, cy - scale)
        line[0].set_gid(self.uuid + '-' + fragment.name.replace(",", '_'))
        annotation_name = re.sub(r'\d,\d', '', fragment.fname)
        if fragment.crossring_cleavages[crossring][1] == "X":
            line = ax.plot((cx + scale, cx + 2 * scale), (cy - scale, cy - scale), color=color, zorder=3)
            line[0].set_gid(self.uuid + '-' + fragment.name + "-direction")
            patch_dict[fragment.name + "_direction"] = line[0]
            self.data['position'][fragment.name + "_direction"] =\
                (cx + scale, cx + 2 * scale), (cy - scale, cy - scale)

            if label:
                ax.text((cx + scale) - 0.4, (cy - scale) - .15, annotation_name)
        else:
            line = ax.plot((cx - scale, cx - scale * 2), (cy + scale, cy + scale), color=color, zorder=3)
            line[0].set_gid(self.uuid + '-' + fragment.name.replace(",", '_') + "-direction")
            patch_dict[fragment.name + "_direction"] = line[0]
            self.data['position'][fragment.name + "_direction"] =\
                (cx - scale, cx - scale * 2), (cy + scale, cy + scale)
            if label:
                ax.text((cx - scale) - 0.32, (cy + scale) + .035, annotation_name)



import unittest
from .common import load, glycoct, glycan, multimap, pickle, named_structures, monosaccharides

from glypy import Substituent, tree
from glypy.structure.fragment import Fragment

Glycan = glycan.Glycan


class GlycanTests(unittest.TestCase):
    _file_path = "./test_data/glycoct.txt"

    def test_from_glycoct(self):
        with open(self._file_path) as stream:
            for structure in glycoct.read(stream):
                self.assertEqual(
                    structure, glycoct.loads(structure.serialize("glycoct")))

    def test_fragments_preserve(self):
        structure = load("branchy_glycan")
        dup = structure.clone()
        self.assertEqual(structure, dup)
        list(dup.fragments('ABY', 2))

        self.assertEqual(structure, dup)

    def test_branch_counts(self):
        structure = load("branchy_glycan")
        self.assertEqual(structure.count_branches(), 3)

    def test_fragments_mass(self):
        structure = load("common_glycan")
        frags = list(structure.fragments('ZCBY', 1))
        import json
        with open('test_data/fragments-example.json') as f:
            frags_ref = json.load(f)
        container = multimap.MultiMap()
        for frag in frags_ref:
            container[frag['kind']] = frag['mass']

        def almost_equal(a, b):
            e = 0.0001
            return (a - e) <= b <= (a + e)

        for frag in frags:
            structure.name_fragment(frag)
            kind = frag.kind
            mass = frag.mass
            candidates = container[kind] + container[kind[::-1]]
            if len(candidates) == 0:
                raise AssertionError("No candidates found for {}".format(frag))
            res = (any(almost_equal(mass, x) for x in candidates))
            if not res:
                raise AssertionError(
                    "{} found no matches in {}".format(frag, candidates))

    def test_reducing_end(self):
        structure = load("common_glycan")
        self.assertEqual(structure.reducing_end, None)
        structure.reducing_end = 1
        self.assertEqual(structure.reducing_end, 1)

    def test_clone(self):
        structure = load("common_glycan")
        ref = structure.clone()
        structure.reducing_end = 1
        self.assertTrue(structure != ref)

    def test_indexing(self):
        structure = load("common_glycan")
        ref = structure.clone()
        for i, node in enumerate(structure.index):
            self.assertEqual(node.id, ref[i].id)
        structure.deindex()
        for i, node in enumerate(structure.index):
            self.assertNotEqual(node.id, ref[i].id)
        structure.index = None
        self.assertRaises(IndexError, lambda: structure[0])

    def test_traversal(self):
        structure = load("common_glycan")
        structure[-
                  1].add_monosaccharide(named_structures.monosaccharides['NeuGc'])
        structure.reindex(method='dfs')
        ref = structure.clone()
        self.assertEqual(structure[-1], ref[-1])
        structure.reindex(method='bfs')
        self.assertNotEqual(structure[-1], ref[-1])

    def test_traversal_by_name(self):
        structure = load("common_glycan")
        structure[-1].add_monosaccharide(
            named_structures.monosaccharides['NeuGc'])
        structure.reindex(method='dfs')
        ref = structure.clone()
        self.assertEqual(structure, ref)
        structure.reindex(method='depth_first_traversal')
        self.assertEqual(structure, ref)
        self.assertRaises(
            KeyError, lambda: structure.reindex(
                method='not_real_traversal'))

    def test_leaves(self):
        structure = load("common_glycan")
        leaves = list(structure.leaves())
        for node in leaves:
            self.assertTrue(len(list(node.children())) == 0)
        leaves = list(structure.leaves(bidirectional=True))
        for node in leaves:
            self.assertTrue(
                len(list(node.children())) == 0 or node == structure.root)

    def test_custom_traversal_method(self):
        def rev_sort_dfs(self, apply_fn=lambda x: x, visited=None, from_node=None, *args, **kwargs):
            node_stack = list([self.root])
            visited = set()
            while len(node_stack) > 0:
                node = node_stack.pop()
                if node.id in visited:
                    continue
                visited.add(node.id)
                yield apply_fn(node)
                node_stack.extend(reversed(list(terminal for pos, link in node.links.items()
                                                for terminal in link if terminal.id not in visited and
                                                len(link.child.substituent_links) < 1)))
        structure = load("common_glycan")
        structure[-3].add_monosaccharide(
            named_structures.monosaccharides['Hex'], 4).add_substituent('methyl', 5)
        structure.reindex()
        ref = structure.clone()
        self.assertEqual(structure[-1], ref[-1])
        structure.reindex(method=rev_sort_dfs)
        self.assertNotEqual(structure[-1], ref[-1])

    def test_topological_equality(self):
        base = load("branchy_glycan")
        a = base.clone()
        b = base.clone()
        c = base.clone()
        self.assertEqual(base, b)
        self.assertEqual(a, b)
        list(a.leaves())[0].add_monosaccharide(monosaccharides["NeuGc"])
        list(b.leaves())[1].add_monosaccharide(monosaccharides["NeuGc"])
        list(c.leaves())[2].add_monosaccharide(monosaccharides["NeuGc"])
        d = base.clone()
        d_children = list(d.leaves())
        d_children[0].add_monosaccharide(monosaccharides["NeuGc"])
        d_children[1].add_monosaccharide(monosaccharides["NeuGc"])
        self.assertTrue(a.topological_equality(b))
        self.assertFalse(a.topological_equality(c))
        self.assertFalse(a.topological_equality(base))
        self.assertFalse(a.topological_equality(d))
        self.assertFalse(b.topological_equality(d))

    def test_substructures_does_not_mutate(self):
        structure = load("broad_n_glycan")
        ref = structure.clone()
        for substructure in structure.substructures(max_cleavages=2):
            pass
        self.assertEqual(structure, ref)

    def test_crossring(self):
        structure = load("branchy_glycan")
        frag_data = {
            '0,2A': 485.174,
            '0,2X': 1317.470,
            '0,3A': 455.163,
            '0,3X': 1347.481,
            '0,4A': 425.153,
            '0,4X': 1377.491,
            '1,3A': 425.153,
            '1,3X': 1377.491,
            '1,4A': 455.163,
            '1,4X': 1347.481,
            '1,5A': 864.322,
            '1,5X': 938.322,
            '2,4X': 1742.623,
            '2,5A': 469.179,
            '2,5X': 1333.465,
            '3,5A': 439.168,
            '3,5X': 1363.4769
        }
        for fragment in structure.fragments("AXB", max_cleavages=2):
            if fragment.link_ids == [3]:
                self.assertAlmostEqual(frag_data[fragment.kind], fragment.mass, 2)

    def test_subtree_from(self):
        structure = load("branchy_glycan")
        child = structure.root.children()[0][1]
        subtree = Glycan.subtree_from(structure, child)
        temp = structure.clone()
        temproot = temp.root.children()[0][1]
        for link in temp.root.links.values():
            link.break_link(refund=True)
        temp.root = temproot
        self.assertEqual(temp, subtree)
        self.assertEqual(Glycan.subtree_from(structure, 1), temp)

    def test_cyclic_clone(self):
        import warnings
        warnings.simplefilter("error")
        with self.assertRaises(UserWarning):
            structure = load("cyclical_glycan")
            self.assertEqual(structure, structure.clone())

    def test_subtree_from_fragment(self):
        structure = load("branchy_glycan")
        fragment = Fragment(mass=1463.5284494, kind="1,5A0,3X", included_nodes=set([2, 3, 4, 5, 6, 7, 8, 9, 10]),
                            link_ids={}, name="0,3Xc4-1,5A4",
                            crossring_cleavages={3: ('1,5', 'A'), 10: ('0,3', 'X')})
        subtree = glycan.fragment_to_substructure(fragment, structure)
        self.assertAlmostEqual(subtree.mass(), fragment.mass)

    def test_get(self):
        structure = load("branchy_glycan")
        self.assertEqual(structure.get(1), structure.root)
        self.assertEqual(structure.get_link(1), next(structure.iterlinks())[1])

    def test_iternodes(self):
        structure = load("branchy_glycan")
        for a, b in zip(iter(structure), structure.iternodes()):
            self.assertEqual(a, b)

    def test_iterlinks_substituents(self):
        had_substituents = False
        structure = load("branchy_glycan")
        for p, link in structure.iterlinks(substituents=True):
            if isinstance(link.child, Substituent):
                had_substituents = True
        self.assertTrue(had_substituents)

    def test_subtrees(self):
        structure = load("branchy_glycan")
        g = structure.subtrees()
        s = next(g)
        self.assertAlmostEqual(tree(s).mass(), 221.0899, 2)
        s = next(g)
        self.assertAlmostEqual(tree(s).mass(), 1599.565, 2)

    def test_fragment_properties(self):
        structure = load("branchy_glycan")
        frags = {f.name: f for f in structure.fragments("Y")}
        self.assertEqual(frags["Ya2"].fname, r"Y$_\alpha$2")
        Ya2 = frags["Ya2"]
        self.assertTrue(Ya2.is_reducing())
        self.assertFalse(Ya2.is_non_reducing())
        self.assertFalse(Ya2.is_internal())

    def test_iterconfigurations(self):
        G81339YK_iupac = (
            '?-D-GalpNAc-(1->3/4)[?-L-Fucp-(1->3/4)]-beta-D-GlcpNAc-(1->3)[?-D-GalpNAc-(1->3/4)[?-L-Fucp-(1->3/4)]'
            '-beta-D-GlcpNAc-(1->6)]-?-D-GalpNAc(1->')
        G81339YK_glycoct = '''RES
1b:x-dgal-HEX-1:5
2s:n-acetyl
3b:b-dglc-HEX-1:5
4b:x-dgal-HEX-1:5
5s:n-acetyl
6b:x-lgal-HEX-1:5|6:d
7s:n-acetyl
8b:b-dglc-HEX-1:5
9b:x-dgal-HEX-1:5
10s:n-acetyl
11b:x-lgal-HEX-1:5|6:d
12s:n-acetyl
LIN
1:1d(2+1)2n
2:1o(3+1)3d
3:3o(3|4+1)4d
4:4d(2+1)5n
5:3o(3|4+1)6d
6:3d(2+1)7n
7:1o(6+1)8d
8:8o(3|4+1)9d
9:9d(2+1)10n
10:8o(3|4+1)11d
11:8d(2+1)12n
'''
        from glypy.io import iupac, glycoct
        structure_iupac = iupac.loads(G81339YK_iupac)
        structure_glycoct = glycoct.loads(G81339YK_glycoct)

        structure = structure_iupac
        configurations = []
        for config_list in structure.iterconfiguration():
            instance = structure.clone()
            for link, conf in config_list:
                link = instance.get_link(link.id)
                parent = instance.get(conf[0].id)
                child = instance.get(conf[1].id)
                link.reconfigure(parent, child, conf[2], conf[3])
            configurations.append(instance)
        configurations_iupac = configurations
        assert len(configurations) == 4

        structure = structure_glycoct
        configurations = []
        for config_list in structure.iterconfiguration():
            instance = structure.clone()
            for link, conf in config_list:
                link = instance.get_link(link.id)
                parent = instance.get(conf[0].id)
                child = instance.get(conf[1].id)
                link.reconfigure(parent, child, conf[2], conf[3])
            configurations.append(instance)
        configurations_glycoct = configurations
        assert len(configurations) == 4

        pairs = zip(sorted(configurations_glycoct, key=str), sorted(configurations_iupac, key=str))
        for a, b in pairs:
            assert a.canonicalize() == b.canonicalize()

    def test_pickles(self):
        ambig = '''
        RES
        1b:x-dglc-HEX-x:x
        2s:n-acetyl
        3b:b-dglc-HEX-1:5
        4s:n-acetyl
        5b:b-dman-HEX-1:5
        6b:a-dman-HEX-1:5
        7b:b-dglc-HEX-1:5
        8s:n-acetyl
        9b:b-dgal-HEX-1:5
        10b:a-dman-HEX-1:5
        11b:a-dman-HEX-1:5
        LIN
        1:1d(2+1)2n
        2:1o(4+1)3d
        3:3d(2+1)4n
        4:3o(4+1)5d
        5:5o(3|6+1)6d
        6:6o(2+1)7d
        7:7d(2+1)8n
        8:7o(4+1)9d
        9:5o(3|6+1)10d
        10:10o(3|6+1)11d
        '''

        x = glycoct.loads(ambig)

        result = pickle.loads(pickle.dumps(x, -1))
        assert x == result

        ambig = """
        RES
        1b:x-dglc-HEX-1:5
        2b:x-dglc-HEX-1:5
        3b:x-dman-HEX-1:5
        4b:a-dman-HEX-1:5
        5b:b-dglc-HEX-1:5
        6s:n-acetyl
        7b:a-dman-HEX-1:5
        8b:b-dglc-HEX-1:5
        9s:n-acetyl
        10b:b-dglc-HEX-1:5
        11b:x-dgal-HEX-1:5
        12b:x-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        13s:n-acetyl
        14s:n-acetyl
        15s:n-acetyl
        16s:n-acetyl
        17b:a-lgal-HEX-1:5|6:d
        LIN
        1:1o(-1+1)2d
        2:2o(-1+1)3d
        3:3o(3+1)4d
        4:4o(2|4+1)5d
        5:5d(2+1)6n
        6:3o(6+1)7d
        7:7o(2+1)8d
        8:8d(2+1)9n
        9:7o(6+1)10d
        10:10o(-1+1)11d
        11:11o(-1+2)12d
        12:12d(5+1)13n
        13:10d(2+1)14n
        14:2d(2+1)15n
        15:1d(2+1)16n
        16:1o(6+1)17d
        """
        x = glycoct.loads(ambig)
        result = pickle.loads(pickle.dumps(x, -1))
        assert x == result

        ambig = """
        RES
        1b:b-dglc-HEX-1:5
        2s:n-acetyl
        3b:b-dglc-HEX-1:5
        4s:n-acetyl
        5b:b-dman-HEX-1:5
        6b:a-dman-HEX-1:5
        7b:b-dglc-HEX-1:5
        8s:n-acetyl
        9b:b-dgal-HEX-1:5
        10b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        11s:n-acetyl
        12b:b-dglc-HEX-1:5
        13s:n-acetyl
        14b:b-dgal-HEX-1:5
        15b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        16s:n-acetyl
        17b:a-dman-HEX-1:5
        18b:b-dglc-HEX-1:5
        19s:n-acetyl
        20b:b-dgal-HEX-1:5
        21b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        22s:n-acetyl
        23b:a-lgal-HEX-1:5|6:d
        LIN
        1:1d(2+1)2n
        2:1o(4+1)3d
        3:3d(2+1)4n
        4:3o(4+1)5d
        5:5o(3|6+1)6d
        6:6o(2+1)7d
        7:7d(2+1)8n
        8:7o(4+1)9d
        9:9o(3|6+2)10d
        10:10d(5+1)11n
        11:6o(4|6+1)12d
        12:12d(2+1)13n
        13:12o(4+1)14d
        14:14o(3|6+2)15d
        15:15d(5+1)16n
        16:5o(3|6+1)17d
        17:17o(2+1)18d
        18:18d(2+1)19n
        19:18o(4+1)20d
        20:20o(3|6+2)21d
        21:21d(5+1)22n
        22:1o(6+1)23d
        """

        x = glycoct.loads(ambig)
        result = pickle.loads(pickle.dumps(x, -1))
        assert x == result

        ambig = """
        RES
        1b:x-dglc-HEX-1:5
        2s:n-acetyl
        3b:b-dglc-HEX-1:5
        4s:n-acetyl
        5b:b-dman-HEX-1:5
        6b:a-dman-HEX-1:5
        7b:b-dglc-HEX-1:5
        8s:n-acetyl
        9b:b-dgal-HEX-1:5
        10b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        11s:n-acetyl
        12b:b-dglc-HEX-1:5
        13s:n-acetyl
        14b:b-dgal-HEX-1:5
        15b:a-dman-HEX-1:5
        16b:b-dglc-HEX-1:5
        17s:n-acetyl
        18b:b-dgal-HEX-1:5
        19b:b-dglc-HEX-1:5
        20s:n-acetyl
        21b:b-dgal-HEX-1:5
        22b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        23s:n-acetyl
        24b:b-dglc-HEX-1:5
        25s:n-acetyl
        26b:b-dgal-HEX-1:5
        27b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
        28s:n-acetyl
        29b:a-lgal-HEX-1:5|6:d
        LIN
        1:1d(2+1)2n
        2:1o(4+1)3d
        3:3d(2+1)4n
        4:3o(4+1)5d
        5:5o(3+1)6d
        6:6o(2+1)7d
        7:7d(2+1)8n
        8:7o(4+1)9d
        9:9o(3|6+2)10d
        10:10d(5+1)11n
        11:6o(4+1)12d
        12:12d(2+1)13n
        13:12o(4+1)14d
        14:5o(6+1)15d
        15:15o(2+1)16d
        16:16d(2+1)17n
        17:16o(4+1)18d
        18:18o(3+1)19d
        19:19d(2+1)20n
        20:19o(4+1)21d
        21:21o(3|6+2)22d
        22:22d(5+1)23n
        23:15o(6+1)24d
        24:24d(2+1)25n
        25:24o(4+1)26d
        26:26o(3|6+2)27d
        27:27d(5+1)28n
        28:1o(6+1)29d
        """
        x = glycoct.loads(ambig)
        result = pickle.loads(pickle.dumps(x, -1))
        assert x == result


if __name__ == '__main__':
    unittest.main()



from cpython.object cimport PyObject
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Next
from cpython.list cimport PyList_Append, PyList_GET_ITEM, PyList_GET_SIZE, PyList_Sort
from cpython.tuple cimport PyTuple_GET_ITEM
from cpython.int cimport PyInt_AsLong, PyInt_FromLong
from glypy.composition.ccomposition cimport CComposition
from glypy.composition import formula

from glypy._c.utils cimport _prepare_glycan_composition_string

cdef CComposition water = CComposition("H2O")


cdef object ZERO = 0


cdef class _CompositionBase(dict):

    @classmethod
    def _empty(cls):
        cdef _CompositionBase inst = cls.__new__(cls)
        inst._mass = None
        inst._reducing_end = None
        inst._composition_offset = CComposition._create(water)
        return inst

    cpdef object _getitem_fast(self, object key):
        cdef:
            PyObject* ptmp

        ptmp = PyDict_GetItem(self, key)
        if ptmp == NULL:
            return ZERO
        return <object>ptmp

    cpdef object _setitem_fast(self, object key, object value):
        PyDict_SetItem(self, key, value)

    cdef void _add_from(self, _CompositionBase other):
        cdef:
            Py_ssize_t i
            PyObject* pkey
            PyObject* pval
            PyObject* ptmp
            long value

        i = 0
        while PyDict_Next(other, &i, &pkey, &pval):
            ptmp = PyDict_GetItem(self, <object>pkey)
            if ptmp == NULL:
                value = 0
            else:
                value = PyInt_AsLong(<object>ptmp)
            value += PyInt_AsLong(<object>pval)
            PyDict_SetItem(self, <object>pkey, PyInt_FromLong(value))

    cdef void _subtract_from(self, _CompositionBase other):
        cdef:
            Py_ssize_t i
            PyObject* pkey
            PyObject* pval
            PyObject* ptmp
            long value

        i = 0
        while PyDict_Next(other, &i, &pkey, &pval):
            ptmp = PyDict_GetItem(self, <object>pkey)
            if ptmp == NULL:
                value = 0
            else:
                value = PyInt_AsLong(<object>ptmp)
            value -= PyInt_AsLong(<object>pval)
            PyDict_SetItem(self, <object>pkey, PyInt_FromLong(value))

    def __reduce__(self):
        return self.__class__, (), self.__getstate__()

    def __getstate__(self):
        d = {
            'mapping': dict(self),
            'reducing_end': self._reducing_end,
            'composition_offset': self._composition_offset
        }
        return d

    def __setstate__(self, state):
        self.update(state['mapping'])
        self._reducing_end = state['reducing_end']
        self._composition_offset = state['composition_offset']

    cpdef _update_from_typed_map(self, _CompositionBase template, bint copy_nodes=False):
        cdef:
            PyObject* pkey
            PyObject* pval
            Py_ssize_t pos
        pos = 0
        if copy_nodes:
            while PyDict_Next(<dict>template, &pos, &pkey, &pval):
                self._setitem_fast((<object>pkey).clone(), <object>pval)
        else:
            while PyDict_Next(<dict>template, &pos, &pkey, &pval):
                self._setitem_fast(<object>pkey, <object>pval)
        reduced = template._reducing_end
        if reduced is not None:
            self._reducing_end = reduced.clone()
        self._mass = None

    cpdef str serialize(self):
        # Causes segmentation fault in unrelated mass calculation code?
        # cdef str result = _prepare_glycan_composition_string(self)
        cdef str result = _reformat(self)
        reduced = self._reducing_end
        if reduced is not None:
            result = "%s$%s" % (result, formula(reduced.total_composition()))
        return result


cdef str _reformat(dict self):
    cdef:
        list tokens, strings
        Py_ssize_t i, n
        PyObject* k
        PyObject* v
        object key
        int count
        tuple token

    tokens = []
    strings = []
    i = 0
    while PyDict_Next(self, &i, &k, &v):
        count = PyInt_AsLong(<object>v)
        if count != 0:
            key = <object>k
            PyList_Append(tokens, (key.mass(), str(key), (<object>v)))
    PyList_Sort(tokens)
    i = 0
    n = PyList_GET_SIZE(tokens)
    for i in range(n):
        token = <tuple>PyList_GET_ITEM(tokens, i)
        name = <object>PyTuple_GET_ITEM(token, 1)
        cnt = <object>PyTuple_GET_ITEM(token, 2)
        part = "%s:%d" % (name, cnt)
        PyList_Append(strings, part)
    return "{%s}" % '; '.join(strings)